import hashlib
from copy import deepcopy
from pathlib import Path
from typing import List

import lxml.etree as ET
import numpy as np
from shapely.geometry import Polygon

from burst2safe.utils import calculate_crc16


SAFE_NS = 'http://www.esa.int/safe/sentinel-1.0'
NAMESPACES = {
    'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
    'gml': 'http://www.opengis.net/gml',
    'xfdu': 'urn:ccsds:schema:xfdu:1',
    'safe': SAFE_NS,
    's1': f'{SAFE_NS}/sentinel-1',
    's1sar': f'{SAFE_NS}/sentinel-1/sar',
    's1sarl1': f'{SAFE_NS}/sentinel-1/sar/level-1',
    's1sarl2': f'{SAFE_NS}/sentinel-1/sar/level-2',
    'gx': 'http://www.google.com/kml/ext/2.2',
}


def get_footprint_string(bbox: Polygon, x_first=True) -> str:
    """Get a string representation of the footprint of the product.

    Args:
        bbox: The bounding box of the product
        x_first: Whether to put the x coordinate first or second

    Returns:
        A string representation of the product footprint
    """
    coords = [(np.round(y, 6), np.round(x, 6)) for x, y in bbox.exterior.coords]
    # TODO: order assumes descending
    coords = [coords[2], coords[3], coords[0], coords[1]]
    if x_first:
        coords_str = ' '.join([f'{x},{y}' for x, y in coords])
    else:
        coords_str = ' '.join([f'{y},{x}' for x, y in coords])
    return coords_str


class Manifest:
    """Class representing a SAFE manifest."""

    def __init__(
        self,
        content_units: List[ET.Element],
        metadata_objects: List[ET.Element],
        data_objects: List[ET.Element],
        bbox: Polygon,
        template_manifest: ET.Element,
    ):
        """Initialize a Manifest object.

        Args:
            content_units: A list of contentUnit elements
            metadata_objects: A list of metadataObject elements
            data_objects: A list of dataObject elements
            bbox: The bounding box of the product
            template_manifest: The template manifest to generate unalterd metadata from
        """
        self.content_units = content_units
        self.metadata_objects = metadata_objects
        self.data_objects = data_objects
        self.bbox = bbox
        self.template = template_manifest
        self.version = 'esa/safe/sentinel-1.0/sentinel-1/sar/level-1/slc/standard/iwdp'

        # Updated by methods
        self.information_package_map = None
        self.metadata_section = None
        self.data_object_section = None
        self.xml = None
        self.path = None
        self.crc = None

    def create_information_package_map(self):
        """Create the information package map."""
        xdfu_ns = NAMESPACES['xfdu']
        information_package_map = ET.Element(f'{{{xdfu_ns}}}informationPackageMap')
        parent_content_unit = ET.Element(
            f'{{{xdfu_ns}}}contentUnit',
            unitType='SAFE Archive Information Package',
            textInfo='Sentinel-1 IW Level-1 SLC Product',
            dmdID='acquisitionPeriod platform generalProductInformation measurementOrbitReference measurementFrameSet',
            pdiID='processing',
        )
        for content_unit in self.content_units:
            parent_content_unit.append(content_unit)

        information_package_map.append(parent_content_unit)
        self.information_package_map = information_package_map

    def create_metadata_section(self):
        """Create the metadata section."""
        metadata_section = ET.Element('metadataSection')
        for metadata_object in self.metadata_objects:
            metadata_section.append(metadata_object)

        ids_to_keep = [
            'processing',
            'platform',
            'measurementOrbitReference',
            'generalProductInformation',
            'acquisitionPeriod',
            'measurementFrameSet',
            's1Level1ProductSchema',
            's1Level1NoiseSchema',
            's1Level1RfiSchema',
            's1Level1CalibrationSchema',
            's1ObjectTypesSchema',
            's1Level1MeasurementSchema',
            's1Level1ProductPreviewSchema',
            's1Level1QuicklookSchema',
            's1MapOverlaySchema',
        ]
        for obj in self.template.find('metadataSection'):
            if obj.get('ID') in ids_to_keep:
                metadata_section.append(deepcopy(obj))

        coordinates = metadata_section.find('.//{*}coordinates')
        coordinates.text = get_footprint_string(self.bbox)

        self.metadata_section = metadata_section

    def create_data_object_section(self):
        """Create the data object section."""
        self.data_object_section = ET.Element('dataObjectSection')
        for data_object in self.data_objects:
            self.data_object_section.append(data_object)

    def assemble(self):
        """Assemble the components of the SAFE manifest."""
        self.create_information_package_map()
        self.create_metadata_section()
        self.create_data_object_section()

        manifest = ET.Element('{%s}XFDU' % NAMESPACES['xfdu'], nsmap=NAMESPACES)
        manifest.set('version', self.version)
        manifest.append(self.information_package_map)
        manifest.append(self.metadata_section)
        manifest.append(self.data_object_section)
        manifest_tree = ET.ElementTree(manifest)

        ET.indent(manifest_tree, space='  ')
        self.xml = manifest_tree

    def write(self, out_path: Path, update_info: bool = True) -> None:
        """Write the SAFE manifest to a file. Optionally update the path and CRC.

        Args:
            out_path: The path to write the manifest to
            update_info: Whether to update the path and CRC
        """
        self.xml.write(out_path, pretty_print=True, xml_declaration=True, encoding='utf-8')
        if update_info:
            self.path = out_path
            self.crc = calculate_crc16(self.path)


class Kml:
    """Class representing a SAFE manifest."""

    def __init__(self, bbox: Polygon):
        """Initialize a KML object.

        Args:
            bbox: The bounding box of the product
        """
        self.bbox = bbox
        self.xml = None

    def assemble(self):
        """Assemble the components of the SAFE KML preview file."""
        kml = ET.Element('kml', nsmap=NAMESPACES)
        document = ET.SubElement(kml, 'Document')
        doc_name = ET.SubElement(document, 'name')
        doc_name.text = 'Sentinel-1 Map Overlay'

        folder = ET.SubElement(document, 'Folder')
        folder_name = ET.SubElement(folder, 'name')
        folder_name.text = 'Sentinel-1 Scene Overlay'

        ground_overlay = ET.SubElement(folder, 'GroundOverlay')
        ground_overlay_name = ET.SubElement(ground_overlay, 'name')
        ground_overlay_name.text = 'Sentinel-1 Image Overlay'
        icon = ET.SubElement(ground_overlay, 'Icon')
        href = ET.SubElement(icon, 'href')
        # TODO: we intentionally don't create this image because we don't know how to.
        href.text = 'quick-look.png'
        lat_lon_quad = ET.SubElement(ground_overlay, f'{{{NAMESPACES["gx"]}}}LatLonQuad')
        coordinates = ET.SubElement(lat_lon_quad, 'coordinates')
        coordinates.text = get_footprint_string(self.bbox, x_first=False)

        kml_tree = ET.ElementTree(kml)
        ET.indent(kml_tree, space='  ')
        self.xml = kml_tree

    def write(self, out_path: Path, update_info: bool = True) -> None:
        """Write the SAFE kml to a file.

        Args:
            out_path: The path to write the manifest to
            update_info: Whether to update the path
        """
        self.xml.write(out_path, pretty_print=True, xml_declaration=True, encoding='utf-8')
        if update_info:
            self.path = out_path
            with open(out_path, 'rb') as f:
                file_bytes = f.read()
                self.size_bytes = len(file_bytes)
                self.md5 = hashlib.md5(file_bytes).hexdigest()

    def update_path(self, safe_path: Path):
        """Update the path based on new a SAFE path.

        Args:
            safe_path: The new SAFE path
        """
        parts = self.path.parts
        parent_index = parts.index(safe_path.parent.name)
        self.path = safe_path / Path(*parts[parent_index + 2 :])


class Preview:
    """Class representing a product preview HTML file."""

    def __init__(
        self,
        name: str,
        product: List[str],
        calibration: List[str],
        measurement: List[str],
        rfi: List[str] = [],
    ):
        """Initialize a Preview object.

        Args:
            name: The name of the product
            product: A list of product annotation files
            calibration: A list of calibration annotation files
            measurement: A list of measurement annotation files
            rfi: A list of rfi annotation files
        """
        self.name = '_'.join(name.split('_')[:-1])
        self.product = product
        self.calibration = calibration
        self.measurement = measurement
        self.rfi = rfi
        self.preview = ['map-overlay.kml', 'product-preview.html', 'quick-look.png']
        self.preview_icon = ['logo.png']
        self.support = [
            's1-level-1-product.xsd',
            's1-level-1-noise.xsd',
            's1-level-1-calibration.xsd',
            's1-object-types.xsd',
            's1-map-overlay.xsd',
            's1-product-preview.xsd',
            's1-level-1-measurement.xsd',
            's1-level-1-quicklook.xsd',
        ]
        if len(self.rfi) > 0:
            self.support.append('s1-level-1-rfi.xsd')
        self.html = None
        self.path = None

    def create_base(self):
        """Create the base HTML product preview."""
        namespaces = {'xsd': 'http://www.w3.org/2001/XMLSchema', 'fn': 'http://www.w3.org/2005/xpath-functions'}
        html = ET.Element('html', nsmap=namespaces)
        head = ET.SubElement(html, 'head')
        ET.SubElement(head, 'meta', attrib={'http-equiv': 'Content-Type', 'content': 'text/html; charset=utf-8'})

        # Create the head section
        nsmap = {'xsd': 'http://www.w3.org/2001/XMLSchema', 'fn': 'http://www.w3.org/2005/xpath-functions'}
        html = ET.Element('html', nsmap=nsmap)
        head = ET.SubElement(html, 'head')

        # Meta element
        ET.SubElement(head, 'meta', attrib={'http-equiv': 'Content-Type', 'content': 'text/html; charset=UTF-8'})

        # Title element
        title = ET.SubElement(head, 'title')
        title.text = self.name

        # Style element
        style = ET.SubElement(head, 'style', attrib={'type': 'text/css'})
        style.text = """
        h1 {font-size:20px}
        h2 {font-size: 18px}
        """

        # Create the body section
        body = ET.SubElement(html, 'body')

        # Add image and title
        ET.SubElement(body, 'img', attrib={'src': 'icons/logo.png'})
        h1 = ET.SubElement(body, 'h1')
        h1.text = self.name

        # Add manifest link
        h2_manifest = ET.SubElement(body, 'h2')
        a_manifest = ET.SubElement(h2_manifest, 'a', attrib={'href': '../manifest.safe'})
        a_manifest.text = 'manifest.safe'

        return html

    def add_subsection(self, body, name, files):
        """Add a file set subsection to the HTML preview."""
        h2 = ET.SubElement(body, 'h2')
        h2.text = name
        ul = ET.SubElement(body, 'ul')
        for file in files:
            li = ET.SubElement(ul, 'li')
            a = ET.SubElement(li, 'a', attrib={'href': f'../{name}/{file}'})
            a.text = file

    def add_img(self, body):
        """Add the image to the HTML preview."""
        ET.SubElement(body, 'img', attrib={'style': 'float:right', 'src': '../preview/quick-look.png'})

    def assemble(self):
        """Assemble the HTML preview."""
        html = self.create_base()
        body = html.find('.//body')

        self.add_subsection(body, 'annotation', self.product)
        self.add_subsection(body, 'annotation/calibration', self.calibration)
        if len(self.rfi) > 0:
            self.add_subsection(body, 'annotation/rfi', self.rfi)
        self.add_subsection(body, 'measurement', self.measurement)
        self.add_img(body)
        self.add_subsection(body, 'preview', self.preview)
        self.add_subsection(body, 'preview/icons', self.preview_icon)
        self.add_subsection(body, 'support', self.support)

        html_tree = ET.ElementTree(html)
        ET.indent(html_tree, space='  ')
        self.html = html_tree

    def write(self, out_path: Path, update_info=True) -> None:
        """Write the html to a file.

        Args:
            out_path: The path to write the annotation to.
            update_info: Whether to update the size and md5 attributes of the html.
        """
        self.html.write(out_path, pretty_print=True, xml_declaration=True, encoding='utf-8')

        if update_info:
            self.path = out_path
            with open(out_path, 'rb') as f:
                file_bytes = f.read()
                self.size_bytes = len(file_bytes)
                self.md5 = hashlib.md5(file_bytes).hexdigest()

    def update_path(self, safe_path: Path):
        """Update the path based on new a SAFE path.

        Args:
            safe_path: The new SAFE path
        """
        parts = self.path.parts
        parent_index = parts.index(safe_path.parent.name)
        self.path = safe_path / Path(*parts[parent_index + 2 :])



from datetime import datetime
from pathlib import Path

import asf_search
import pytest

from burst2safe.utils import BurstInfo


TEST_DIR = Path(__file__).parent


def pytest_configure(config):
    config.addinivalue_line('filterwarnings', 'ignore::RuntimeWarning')


@pytest.fixture
def test_data_dir():
    return TEST_DIR / 'test_data'


# @pytest.fixture
# def test_data_xml(test_data_dir):
#     return test_data_dir / 'S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85_VV.xml'


@pytest.fixture
def test_data1_xml(test_data_dir):
    return test_data_dir / 'S1A_IW_SLC__1SDV_20240408T015045_20240408T015113_053336_06778C_CB5D_VV.xml'


@pytest.fixture
def test_data2_xml(test_data_dir):
    return test_data_dir / 'S1A_IW_SLC__1SDV_20240408T015111_20240408T015138_053336_06778C_CA9A_VV.xml'


@pytest.fixture
def xsd_dir():
    xsd_dir = Path(__file__).parent.parent / 'src' / 'burst2safe' / 'data' / 'support_340'
    return xsd_dir


@pytest.fixture
def search_result1():
    product = asf_search.ASFProduct()
    product.umm = {'InputGranules': ['S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85']}
    product.properties.update(
        {
            'fileID': 'S1_136231_IW2_20200604T022312_VV_7C85-BURST',
            'flightDirection': 'ascending',
            'polarization': 'vv',
            'orbit': 123,
            'pathNumber': 456,
            'url': 'https://example.com/foo.zip',
            'additionalUrls': ['https://example.com/foo.xml'],
            'burst': {
                'subswath': 'IW2',
                'relativeBurstID': 123456,
                'burstIndex': 7,
            },
        }
    )
    # If an ASFSearchResults object is needed, uncomment the following lines:
    # results = asf_search.ASFSearchResults([product])
    # results.searchComplete = True
    return product


@pytest.fixture
def burst_info1(test_data1_xml):
    burst_info = BurstInfo(
        granule='S1_135526_IW2_20240408T015108_VV_CB5D-BURST',
        slc_granule='S1A_IW_SLC__1SDV_20240408T015045_20240408T015113_053336_06778C_CB5D',
        swath='IW2',
        polarization='VV',
        burst_id=135526,
        burst_index=8,
        direction='ASCENDING',
        absolute_orbit=53336,
        relative_orbit=64,
        date=datetime(2024, 4, 8, 1, 51, 8),
        data_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20240408T015045_20240408T015113_053336_06778C_CB5D/IW2/VV/8.tiff',
        data_path=Path(),
        metadata_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20240408T015045_20240408T015113_053336_06778C_CB5D/IW2/VV/8.xml',
        metadata_path=test_data1_xml,
        start_utc=datetime(2024, 4, 8, 1, 51, 8, 355601),
        stop_utc=datetime(2024, 4, 8, 1, 51, 11, 453324),
        length=1508,
        width=25470,
    )
    return burst_info


@pytest.fixture
def burst_info2(test_data2_xml):
    burst_info = BurstInfo(
        granule='S1_135527_IW2_20240408T015111_VV_CA9A-BURST',
        slc_granule='S1A_IW_SLC__1SDV_20240408T015111_20240408T015138_053336_06778C_CA9A',
        swath='IW2',
        polarization='VV',
        burst_id=135527,
        burst_index=0,
        direction='ASCENDING',
        absolute_orbit=53336,
        relative_orbit=64,
        date=datetime(2024, 4, 8, 1, 51, 11),
        data_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20240408T015111_20240408T015138_053336_06778C_CA9A/IW2/VV/0.tiff',
        data_path=Path(),
        metadata_url='https://sentinel1-burst.asf.alaska.edu/S1A_IW_SLC__1SDV_20240408T015111_20240408T015138_053336_06778C_CA9A/IW2/VV/0.xml',
        metadata_path=test_data2_xml,
        start_utc=datetime(2024, 4, 8, 1, 51, 11, 107991),
        stop_utc=datetime(2024, 4, 8, 1, 51, 14, 205714),
        length=1508,
        width=25470,
    )
    return burst_info


@pytest.fixture
def burst_infos(burst_info1, burst_info2):
    return [burst_info1, burst_info2]



import os
import shutil
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile

import asf_search as asf
import lxml.etree as ET
import numpy as np

from burst2safe.utils import get_burst_infos


VERSIONS = [
    ('2.36', datetime(2014, 10, 1), True),
    ('2.43', datetime(2015, 3, 19), False),
    ('2.45', datetime(2015, 6, 17), True),
    ('2.52', datetime(2015, 7, 2), False),
    ('2.53', datetime(2015, 7, 18), False),  # may not need
    ('2.60', datetime(2015, 11, 20), True),
    ('2.62', datetime(2016, 3, 14), False),
    ('2.70', datetime(2016, 4, 13), False),
    ('2.71', datetime(2016, 5, 11), False),
    ('2.72', datetime(2016, 8, 23), False),
    ('2.82', datetime(2017, 3, 28), False),
    ('2.84', datetime(2017, 8, 22), False),  # may not need
    ('2.90', datetime(2018, 3, 13), True),
    ('2.91', datetime(2018, 6, 26), False),
    ('3.10', datetime(2019, 6, 26), False),
    ('3.20', datetime(2020, 1, 29), False),
    ('3.30', datetime(2020, 6, 23), False),  # may not need
    ('3.31', datetime(2020, 6, 30), False),
    ('3.40', datetime(2021, 11, 4), True),
    ('3.51', datetime(2022, 3, 23), False),
    ('3.52', datetime(2022, 5, 12), False),  # may not need
    ('3.61', datetime(2023, 3, 30), False),
    ('3.71', datetime(2023, 10, 19), True),
    ('Current', datetime.now(), False),
]


def find_representative_bursts(important_only=False):
    options = {
        'intersectsWith': 'POLYGON((12.2376 41.741,12.2607 41.741,12.2607 41.7609,12.2376 41.7609,12.2376 41.741))',
        'dataset': 'SLC-BURST',
        'relativeOrbit': 117,
        'flightDirection': 'Ascending',
        'polarization': 'VV',
        'maxResults': 2000,
    }
    results = asf.search(**options)
    bursts = sorted(get_burst_infos(results, Path.cwd()), key=lambda x: x.date)
    mid_bursts = []
    for i in range(len(VERSIONS) - 1):
        date1 = VERSIONS[i][1]
        date2 = VERSIONS[i + 1][1]
        bursts_between = [burst for burst in bursts if date1 <= burst.date < date2]
        mid_index = int(np.floor(len(bursts_between) / 2))
        mid_burst = bursts_between[mid_index]
        if important_only:
            if VERSIONS[i][2]:
                mid_bursts.append(mid_burst)
        else:
            mid_bursts.append(mid_burst)
    return mid_bursts


def download_slcs():
    slcs = [f'{burst.slc_granule}-SLC' for burst in find_representative_bursts()]
    slcs = asf.granule_search(slcs)
    slcs.download('.')


def get_version(slc_path):
    slc_name = f"{slc_path.name.split('.')[0]}.SAFE"
    with ZipFile(slc_path) as z:
        manifest_str = z.read(f'{slc_name}/manifest.safe')
        manifest = ET.fromstring(manifest_str)

    version_xml = [elem for elem in manifest.findall('.//{*}software') if elem.get('name') == 'Sentinel-1 IPF'][0]
    return version_xml.get('version')


def get_versions(slc_paths):
    versions = [(slc_path.name, get_version(slc_path)) for slc_path in slc_paths]
    versions.sort(key=lambda x: x[1])


def extract_support_folder(slc_path):
    version = get_version(slc_path).replace('.', '')
    out_dir = Path(f'support_{version}')
    out_dir.mkdir(exist_ok=True)
    slc_name = f"{slc_path.name.split('.')[0]}.SAFE"
    with ZipFile(slc_path) as zip_ref:
        for file_info in zip_ref.infolist():
            if file_info.filename.startswith(f'{slc_name}/support/') and not file_info.is_dir():
                source_file = zip_ref.open(file_info)
                target_path = out_dir / Path(file_info.filename).name
                with open(target_path, 'wb') as target_file:
                    shutil.copyfileobj(source_file, target_file)


def create_diffs():
    supports = sorted(Path().glob('support*'))
    for i in range(len(supports) - 1):
        support1 = supports[i]
        support2 = supports[i + 1]
        diff_file = Path(f'diff_{support1.name}_{support2.name}.txt')
        diff_file.touch()
        os.system(f'git diff --no-index {support1} {support2} > {diff_file}')


def identify_changing_versions():
    download_slcs()
    slc_paths = sorted(list(Path().glob('*.zip')))
    for slc_path in slc_paths:
        extract_support_folder(slc_path)
    create_diffs()


def download_changing_metadata():
    bursts = find_representative_bursts()
    sorted_bursts = sorted(bursts, key=lambda x: x.date)

    important_versions = []
    for (version, _, important), burst in zip(VERSIONS, sorted_bursts):
        if important:
            print(version, burst.granule)
            important_versions.append(burst)

    for burst in important_versions:
        asf.download_url(burst.metadata_url, path='.', filename=burst.metadata_path.name)


def download_representative_support():
    slcs = [f'{burst.slc_granule}-SLC' for burst in find_representative_bursts(important_only=True)]
    slcs = asf.granule_search(slcs)
    slcs.download('.')
    slc_paths = sorted(list(Path().glob('*.zip')))
    for slc_path in slc_paths:
        extract_support_folder(slc_path)


if __name__ == '__main__':
    # identify_changing_versions()
    # download_changing_metadata()
    download_representative_support()



from collections.abc import Iterable
from datetime import datetime
from pathlib import Path

from shapely.geometry import MultiPoint, Polygon

from burst2safe.calibration import Calibration
from burst2safe.measurement import Measurement
from burst2safe.noise import Noise
from burst2safe.product import Product
from burst2safe.rfi import Rfi
from burst2safe.utils import BurstInfo


class Swath:
    """Class representing a single swath (and polarization) of a SAFE file."""

    def __init__(
        self,
        burst_infos: Iterable[BurstInfo],
        safe_path: Path,
        version: str,
        creation_time: datetime,
        image_number: int,
    ):
        """Initialize a Swath object.

        Args:
            burst_infos: A list of BurstInfo objects
            safe_path: The path to the SAFE directory
            version: The IPF version of the SAFE file
            creation_time: The creation time of the SAFE file
            image_number: The image number of the swath
        """
        self.check_burst_group_validity(burst_infos)
        self.burst_infos = sorted(burst_infos, key=lambda x: x.burst_id)
        self.safe_path = safe_path
        self.version = version
        self.creation_time = creation_time
        self.image_number = image_number
        self.swath = self.burst_infos[0].swath
        self.polarization = self.burst_infos[0].polarization

        self.name = self.get_swath_name(self.burst_infos, self.safe_path, self.image_number)
        self.major_version, self.minor_version = [int(x) for x in self.version.split('.')]

        self.measurement_name = self.safe_path / 'measurement' / f'{self.name}.tiff'
        self.product_name = self.safe_path / 'annotation' / f'{self.name}.xml'
        self.noise_name = self.safe_path / 'annotation' / 'calibration' / f'noise-{self.name}.xml'
        self.calibration_name = self.safe_path / 'annotation' / 'calibration' / f'calibration-{self.name}.xml'

        self.rfi_name = None
        self.has_rfi = False
        if self.major_version >= 3 and self.minor_version >= 40:
            self.has_rfi = True
            self.rfi_name = self.safe_path / 'annotation' / 'rfi' / f'rfi-{self.name}.xml'

        # Set on write
        self.bbox = None

    @staticmethod
    def check_burst_group_validity(burst_infos: Iterable[BurstInfo]):
        """Check that the burst group is valid.

        The burst group must:
        - Not contain duplicate granules
        - Have the same absolute orbit
        - Be from the same swath
        - Have the same polarization
        - Have consecutive burst IDs

        Args:
            burst_infos: A list of BurstInfo objects
        """
        granules = [x.granule for x in burst_infos]
        duplicates = list(set([x for x in granules if granules.count(x) > 1]))
        if duplicates:
            raise ValueError(f'Found duplicate granules: {duplicates}.')

        orbits = set([x.absolute_orbit for x in burst_infos])
        if len(orbits) != 1:
            raise ValueError(f'All bursts must have the same absolute orbit. Found: {orbits}.')

        swaths = set([x.swath for x in burst_infos])
        if len(swaths) != 1:
            raise ValueError(f'All bursts must be from the same swath. Found: {swaths}.')

        polarizations = set([x.polarization for x in burst_infos])
        if len(polarizations) != 1:
            raise ValueError(f'All bursts must have the same polarization. Found: {polarizations}.')

        burst_ids = [x.burst_id for x in burst_infos]
        burst_ids.sort()
        if burst_ids != list(range(min(burst_ids), max(burst_ids) + 1)):
            raise ValueError(f'All bursts must have consecutive burst IDs. Found: {burst_ids}.')

    @staticmethod
    def get_swath_name(burst_infos: Iterable[BurstInfo], safe_path: Path, image_number: int) -> str:
        """Get the name of the swath. Will be used to name constituent output files.

        Args:
            burst_infos: A list of BurstInfo objects
            safe_path: The path to the SAFE directory
            image_number: The image number of the swath

        Returns:
            The name of the swath
        """
        swath = burst_infos[0].swath.lower()
        pol = burst_infos[0].polarization.lower()
        start = datetime.strftime(min([x.start_utc for x in burst_infos]), '%Y%m%dt%H%M%S')
        stop = datetime.strftime(max([x.stop_utc for x in burst_infos]), '%Y%m%dt%H%M%S')

        safe_name = safe_path.name
        platfrom, _, _, _, _, _, _, orbit, data_take, _ = safe_name.lower().split('_')
        swath_name = f'{platfrom}-{swath}-slc-{pol}-{start}-{stop}-{orbit}-{data_take}-{image_number:03d}'
        return swath_name

    def get_bbox(self):
        """Get the bounding box of the swath using the Product GCPs.

        Returns:
            The bounding box of the swath
        """
        points = MultiPoint([(gcp.x, gcp.y) for gcp in self.product.gcps])
        min_rotated_rect = points.minimum_rotated_rectangle
        bbox = Polygon(min_rotated_rect.exterior)
        return bbox

    def assemble(self):
        """Assemble the components of the Swath."""
        self.product = Product(self.burst_infos, self.version, self.image_number)
        self.noise = Noise(self.burst_infos, self.version, self.image_number)
        self.calibration = Calibration(self.burst_infos, self.version, self.image_number)
        self.annotations = [self.product, self.noise, self.calibration]

        if self.has_rfi:
            self.rfi = Rfi(self.burst_infos, self.version, self.image_number)
            self.annotations.append(self.rfi)

        for component in self.annotations:
            component.assemble()

        self.measurement = Measurement(
            self.burst_infos, self.product.gcps, self.creation_time, self.version, self.image_number
        )

    def write(self, update_info: bool = True):
        """Write the Swath componets to the SAFE directory.

        Args:
            update_info: Whether to update the bounding box of the Swath
        """
        self.measurement.write(self.measurement_name)
        self.product.update_data_stats(self.measurement.data_mean, self.measurement.data_std)
        self.product.update_burst_byte_offsets(self.measurement.byte_offsets)
        self.product.write(self.product_name)
        self.noise.write(self.noise_name)
        self.calibration.write(self.calibration_name)
        if self.has_rfi:
            self.rfi.write(self.rfi_name)

        if update_info:
            self.bbox = self.get_bbox()

    def update_paths(self, safe_path: Path):
        """Update the paths of the Swath components based on a SAFE path.

        Args:
            safe_path: The new SAFE path
        """
        for component in self.annotations:
            parts = component.path.parts
            parent_index = parts.index(safe_path.parent.name)
            component.path = safe_path / Path(*parts[parent_index + 2 :])

        parts = self.measurement.path.parts
        parent_index = parts.index(safe_path.parent.name)
        self.measurement.path = safe_path / Path(*parts[parent_index + 2 :])

    def create_manifest_components(self):
        """Create the manifest components for the Swath."""
        self.manifest_components = {
            'content_unit': [],
            'metadata_object': [],
            'data_object': [],
            'content_unit_measurement': [],
            'data_object_measurement': [],
        }
        for annotation in self.annotations:
            content, metadata, data = annotation.create_manifest_components()
            self.manifest_components['content_unit'].append(content)
            self.manifest_components['metadata_object'].append(metadata)
            self.manifest_components['data_object'].append(data)

        content, data = self.measurement.create_manifest_components()
        self.manifest_components['content_unit_measurement'].append(content)
        self.manifest_components['data_object_measurement'].append(data)



"""A tool for converting stacks of ASF burst SLCs to stacks of SAFEs"""

from argparse import ArgumentParser
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path
from typing import Optional

from shapely.geometry import Polygon

from burst2safe import utils
from burst2safe.burst2safe import burst2safe
from burst2safe.search import find_stack_orbits


DESCRIPTION = """Convert a stack of ASF burst SLCs to a stack of ESA SAFEs.
This will produce a SAFE for each absolute orbit in the stack. You can
define a stack by specifying a relaitve orbit number, start/end dates,
an extent, and polarization(s).
"""


def burst2stack(
    rel_orbit: Optional[int] = None,
    start_date: Optional[datetime] = None,
    end_date: Optional[datetime] = None,
    extent: Optional[Polygon] = None,
    polarizations: Optional[Iterable[str]] = None,
    swaths: Optional[Iterable[str]] = None,
    mode: str = 'IW',
    min_bursts: int = 1,
    all_anns: bool = False,
    keep_files: bool = False,
    work_dir: Optional[Path] = None,
) -> Path:
    """Convert a stack of burst granules to a stack of ESA SAFEs.
    Wraps the burst2safe function to handle multiple dates.

    Args:
        rel_orbit: The relative orbit number of the bursts
        start_date: The start date of the bursts
        end_date: The end date of the bursts
        extent: The bounding box of the bursts
        swaths: List of swaths to include
        polarizations: List of polarizations to include
        mode: The collection mode to use (IW or EW) (default: IW)
        min_bursts: The minimum number of bursts per swath (default: 1)
        all_anns: Include product annotation files for all swaths, regardless of included bursts
        keep_files: Keep the intermediate files
        work_dir: The directory to create the SAFE in (default: current directory)
    """
    absolute_orbits = find_stack_orbits(rel_orbit, extent, start_date, end_date)
    print(f'Creating SAFEs for {len(absolute_orbits)} time periods...')
    for orbit in absolute_orbits:
        print()
        burst2safe(
            granules=None,
            orbit=orbit,
            extent=extent,
            polarizations=polarizations,
            swaths=swaths,
            mode=mode,
            min_bursts=min_bursts,
            all_anns=all_anns,
            keep_files=keep_files,
            work_dir=work_dir,
        )


def main() -> None:
    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--rel-orbit', type=int, help='Relative orbit number of the bursts')
    parser.add_argument('--start-date', type=str, help='Start date of the bursts (YYYY-MM-DD)')
    parser.add_argument('--end-date', type=str, help='End date of the bursts (YYYY-MM-DD)')
    parser.add_argument(
        '--extent',
        type=str,
        nargs='+',
        help='Bounds (W S E N in lat/lon) or geometry file describing spatial extent',
    )
    parser.add_argument('--pols', type=str, nargs='+', help='Plarizations of the bursts (i.e., VV VH)')
    parser.add_argument('--swaths', type=str, nargs='+', help='Swaths of the bursts (i.e., IW1 IW2 IW3)')
    parser.add_argument('--mode', type=str, default='IW', help='Collection mode to use (IW or EW). Default: IW')
    parser.add_argument('--min-bursts', type=int, default=1, help='Minimum # of bursts per swath/polarization.')
    parser.add_argument(
        '--all-anns',
        action='store_true',
        default=False,
        help='Include product annotations files for all swaths, regardless of included bursts.',
    )
    parser.add_argument('--keep-files', action='store_true', default=False, help='Keep the intermediate files')
    parser.add_argument('--output-dir', type=str, default=None, help='Output directory to save to')

    args = utils.reparse_args(parser.parse_args(), tool='burst2stack')

    burst2stack(
        rel_orbit=args.rel_orbit,
        start_date=args.start_date,
        end_date=args.end_date,
        extent=args.extent,
        polarizations=args.pols,
        swaths=args.swaths,
        mode=args.mode,
        min_bursts=args.min_bursts,
        all_anns=args.all_anns,
        keep_files=args.keep_files,
        work_dir=args.output_dir,
    )



import shutil
from pathlib import Path

import asf_search
import pytest

from burst2safe.safe import Safe
from burst2safe.utils import get_burst_infos
from helpers import create_test_geotiff, validate_xml


BURSTS = [
    ('2.36', 'S1_249405_IW2_20150117T170523_VV_AE35-BURST'),
    ('2.45', 'S1_249405_IW2_20150622T170528_VV_EFBC-BURST'),
    ('2.60', 'S1_249405_IW2_20160124T170529_VV_7CC5-BURST'),
    ('2.90', 'S1_249405_IW2_20180507T170501_VV_6B12-BURST'),
    ('3.40', 'S1_249405_IW2_20211223T170608_VV_B8E2-BURST'),
    ('3.71', 'S1_249405_IW2_20240130T170617_VV_9A35-BURST'),
]


@pytest.mark.integration()
@pytest.mark.parametrize('version,burst', BURSTS)
def test_versions(version, burst, tmp_path):
    products = asf_search.granule_search([burst])
    burst_infos = get_burst_infos(products, tmp_path)

    ipf_dir = Path(__file__).parent / 'test_data' / 'ipf'
    shutil.copy(ipf_dir / burst_infos[0].metadata_path.name, burst_infos[0].metadata_path)

    [info.add_shape_info() for info in burst_infos]
    [info.add_start_stop_utc() for info in burst_infos]

    create_test_geotiff(burst_infos[0].data_path, 'cfloat', shape=(burst_infos[0].length, burst_infos[0].width, 1))
    safe = Safe(burst_infos, work_dir=tmp_path)
    safe.create_safe()

    xsd_dir = Path(__file__).parent.parent / 'src' / 'burst2safe' / 'data' / f'support_{version.replace(".", "")}'
    support_source_dir = safe.support_dir
    assert xsd_dir.name == support_source_dir.name

    support_dir = safe.safe_path / 'support'

    validate_xml(safe.kml.path, support_dir / 's1-map-overlay.xsd')
    validate_xml(safe.preview.path, support_dir / 's1-product-preview.xsd')
    for swath in safe.swaths:
        validate_xml(swath.product.path, support_dir / 's1-level-1-product.xsd')
        validate_xml(swath.noise.path, support_dir / 's1-level-1-noise.xsd')
        validate_xml(swath.calibration.path, support_dir / 's1-level-1-calibration.xsd')
        if swath.has_rfi:
            validate_xml(swath.rfi.path, support_dir / 's1-level-1-rfi.xsd')



import bisect
import shutil
from collections.abc import Iterable
from datetime import datetime
from itertools import product
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
from shapely.geometry import MultiPolygon, Polygon

from burst2safe.base import create_content_unit, create_data_object, create_metadata_object
from burst2safe.manifest import Kml, Manifest, Preview
from burst2safe.product import Product
from burst2safe.swath import Swath
from burst2safe.utils import BurstInfo, drop_duplicates, flatten, get_subxml_from_metadata, optional_wd


class Safe:
    """Class representing a SAFE file."""

    def __init__(self, burst_infos: Iterable[BurstInfo], all_anns: bool = False, work_dir: Optional[Path] = None):
        """Initialize a Safe object.

        Args:
            burst_infos: A list of BurstInfo objects
            all_anns: Include product annotation files for all swaths, regardless of included bursts
            work_dir: The directory to create the SAFE in
        """
        self.burst_infos = burst_infos
        self.all_anns = all_anns
        self.work_dir = optional_wd(work_dir)

        self.check_group_validity(self.burst_infos)

        self.grouped_burst_infos = self.group_burst_infos(self.burst_infos)
        self.name = self.get_name()
        self.safe_path = self.work_dir / self.name
        self.swaths = []
        self.blank_products = []
        self.manifest = None
        self.kml = None

        self.version = self.get_ipf_version(self.burst_infos[0].metadata_path)
        self.major_version, self.minor_version = [int(x) for x in self.version.split('.')]
        self.support_dir = self.get_support_dir()
        self.creation_time = self.get_creation_time()

    def get_creation_time(self) -> datetime:
        """Get the creation time of the SAFE file.
        Always set to the latest SLC processing stop time.

        Returns:
            The creation time of the SAFE file
        """
        metadata_paths = list(set([x.metadata_path for x in self.burst_infos]))
        manifests = [get_subxml_from_metadata(metadata_path, 'manifest') for metadata_path in metadata_paths]
        manifest = manifests[0]
        desired_tag = './/{http://www.esa.int/safe/sentinel-1.0}processing'
        creation_times = []
        for manifest in manifests:
            slc_processing = [elem for elem in manifest.findall(desired_tag) if elem.get('name') == 'SLC Processing'][0]
            creation_times.append(datetime.strptime(slc_processing.get('stop'), '%Y-%m-%dT%H:%M:%S.%f'))
        creation_time = max(creation_times)
        return creation_time

    def get_support_dir(self) -> Path:
        """Find the support directory version closest to but not exceeding the IPF major.minor verion"""
        data_dir = Path(__file__).parent / 'data'
        support_dirs = sorted([x for x in data_dir.iterdir() if x.is_dir()])
        support_versions = sorted([int(x.name.split('_')[1]) for x in support_dirs])
        safe_version = (self.major_version * 100) + self.minor_version

        if safe_version in support_versions:
            support_version = safe_version
        else:
            support_version = support_versions[bisect.bisect_left(support_versions, safe_version) - 1]

        return data_dir / f'support_{support_version}'

    @staticmethod
    def check_group_validity(burst_infos: Iterable[BurstInfo]) -> None:
        """Check that the burst group is valid.

        A valid burst group must:
        - Have the same acquisition mode
        - Be from the same absolute orbit
        - Be contiguous in time and space
        - Have the same footprint for all included polarizations

        Args:
            burst_infos: A list of BurstInfo objects
        """
        swaths = sorted(list(set([info.swath for info in burst_infos])))
        polarizations = sorted(list(set([info.polarization for info in burst_infos])))
        burst_range = {}
        for swath in swaths:
            burst_range[swath] = {}
            for pol in polarizations:
                burst_subset = [info for info in burst_infos if info.swath == swath and info.polarization == pol]
                if len(burst_subset) == 0:
                    burst_range[swath][pol] = [0, 0]
                    continue
                Swath.check_burst_group_validity(burst_subset)

                burst_ids = [info.burst_id for info in burst_subset]
                burst_range[swath][pol] = [min(burst_ids), max(burst_ids)]

            start_ids = [id_range[0] for id_range in burst_range[swath].values()]
            if len(set(start_ids)) != 1:
                raise ValueError(
                    f'Polarization groups in swath {swath} do not have same start burst id. Found {start_ids}'
                )

            end_ids = [id_range[1] for id_range in burst_range[swath].values()]
            if len(set(end_ids)) != 1:
                raise ValueError(f'Polarization groups in swath {swath} do not have same end burst id. Found {end_ids}')

        if len(swaths) == 1:
            return

        swath_combos = [[swaths[i], swaths[i + 1]] for i in range(len(swaths) - 1)]
        working_pol = polarizations[0]
        for swath1, swath2 in swath_combos:
            min_diff = burst_range[swath1][working_pol][0] - burst_range[swath2][working_pol][0]
            if np.abs(min_diff) > 1:
                raise ValueError(f'Products from swaths {swath1} and {swath2} do not overlap')
            max_diff = burst_range[swath1][working_pol][1] - burst_range[swath2][working_pol][1]
            if np.abs(max_diff) > 1:
                raise ValueError(f'Products from swaths {swath1} and {swath2} do not overlap')

    def get_name(self, unique_id: str = '0000') -> str:
        """Create a name for the SAFE file.

        Args:
            burst_infos: A list of BurstInfo objects
            unique_id: A unique identifier for the SAFE file

        Returns:
            The name of the SAFE file
        """
        platform, beam_mode, product_type = self.burst_infos[0].slc_granule.split('_')[:3]

        pol_codes = {'HH': 'SH', 'VV': 'SV', 'VH': 'SV', 'HV': 'SV', 'HH_HV': 'DH', 'VH_VV': 'DV'}
        pols = sorted(list(set([x.polarization for x in self.burst_infos])))
        pol_code = pol_codes['_'.join(pols)]
        product_info = f'1S{pol_code}'

        min_date = min([x.date for x in self.burst_infos]).strftime('%Y%m%dT%H%M%S')
        max_date = max([x.date for x in self.burst_infos]).strftime('%Y%m%dT%H%M%S')
        absolute_orbit = f'{self.burst_infos[0].absolute_orbit:06d}'
        mission_data_take = self.burst_infos[0].slc_granule.split('_')[-2]
        product_name = f'{platform}_{beam_mode}_{product_type}__{product_info}_{min_date}_{max_date}_{absolute_orbit}_{mission_data_take}_{unique_id}.SAFE'
        return product_name

    @staticmethod
    def group_burst_infos(burst_infos: Iterable[BurstInfo]) -> dict:
        """Group burst infos by swath and polarization.

        Args:
            burst_infos: A list of BurstInfo objects

        Returns:
            A dictionary of burst infos grouped by swath, then polarization
        """
        burst_dict = {}
        for burst_info in burst_infos:
            if burst_info.swath not in burst_dict:
                burst_dict[burst_info.swath] = {}

            if burst_info.polarization not in burst_dict[burst_info.swath]:
                burst_dict[burst_info.swath][burst_info.polarization] = []

            burst_dict[burst_info.swath][burst_info.polarization].append(burst_info)

        swaths = list(burst_dict.keys())
        polarizations = list(burst_dict[swaths[0]].keys())
        for swath, polarization in zip(swaths, polarizations):
            burst_dict[swath][polarization] = sorted(burst_dict[swath][polarization], key=lambda x: x.burst_id)

        return burst_dict

    @staticmethod
    def get_ipf_version(metadata_path: Path) -> str:
        """Get the IPF version from the parent manifest file.

        Returns:
            The IPF version as a string
        """
        manifest = get_subxml_from_metadata(metadata_path, 'manifest')
        version_xml = [elem for elem in manifest.findall('.//{*}software') if elem.get('name') == 'Sentinel-1 IPF'][0]
        return version_xml.get('version')

    def get_bbox(self) -> Polygon:
        """Get the bounding box for the SAFE file.

        Returns:
            A Polygon object representing the bounding box
        """
        bboxs = MultiPolygon([swath.bbox for swath in self.swaths])
        min_rotated_rect = bboxs.minimum_rotated_rectangle
        bbox = Polygon(min_rotated_rect.exterior)
        return bbox

    def create_dir_structure(self) -> Path:
        """Create a directory for the SAFE file.

        Returns:
            The path to the SAFE directory
        """
        measurements_dir = self.safe_path / 'measurement'
        annotations_dir = self.safe_path / 'annotation'
        preview_dir = self.safe_path / 'preview'
        icon_dir = preview_dir / 'icons'
        calibration_dir = annotations_dir / 'calibration'
        rfi_dir = annotations_dir / 'rfi'

        calibration_dir.mkdir(parents=True, exist_ok=True)
        measurements_dir.mkdir(parents=True, exist_ok=True)
        icon_dir.mkdir(parents=True, exist_ok=True)
        if self.major_version >= 3 and self.minor_version >= 40:
            rfi_dir.mkdir(parents=True, exist_ok=True)

        shutil.copytree(self.support_dir, self.safe_path / 'support', dirs_exist_ok=True)
        shutil.copy(self.support_dir.parent / 'logo.png', icon_dir / 'logo.png')

    @staticmethod
    def create_representative_burst_set(template_bursts: Iterable[BurstInfo], swath: str, pol: str) -> List[BurstInfo]:
        """Create a representative burst set for a blank product.

        Args:
            template_bursts: A list of BurstInfo objects
            swath: The swath of the blank product
            pol: The polarization of the blank product
        """
        unique_slcs = list(set([x.slc_granule for x in template_bursts]))
        representative_bursts = []
        for slc in unique_slcs:
            slc_bursts = [x for x in template_bursts if x.slc_granule == slc]
            start_utc = min([x.start_utc for x in slc_bursts])
            stop_utc = max([x.stop_utc for x in slc_bursts])
            slc_template = slc_bursts[0]
            new_burst = BurstInfo(
                None,
                None,
                swath,
                pol,
                None,
                0,
                slc_template.direction,
                slc_template.absolute_orbit,
                slc_template.relative_orbit,
                None,
                None,
                None,
                None,
                slc_template.metadata_path,
                start_utc,
                stop_utc,
            )
            new_burst.add_shape_info()
            representative_bursts.append(new_burst)
        return representative_bursts

    def create_blank_products(self, image_number: int) -> List[Product]:
        """Create blank product annotation for missing swaths.

        Args:
            image_number: The starting image number for the annotation products

        Returns:
            A list of blank Product objects
        """
        swaths = list(set([burst.swath for burst in self.burst_infos]))
        missing_swaths = list(set(['IW1', 'IW2', 'IW3']) - set(swaths))
        if not self.all_anns or len(missing_swaths) == 0:
            return []

        pols = list(set([burst.polarization for burst in self.burst_infos]))

        blank_products = []
        for swath, pol in product(missing_swaths, pols):
            image_number += 1
            relevant_bursts = flatten([self.grouped_burst_infos[s][pol] for s in swaths])
            rep_bursts = self.create_representative_burst_set(relevant_bursts, swath, pol)
            annotation = Product(rep_bursts, self.version, image_number, dummy=True)
            blank_products.append(annotation)
        return blank_products

    def create_safe_components(self) -> None:
        """Create the components (data and metadata files) of the SAFE file."""
        swaths = list(self.grouped_burst_infos.keys())
        polarizations = list(self.grouped_burst_infos[swaths[0]].keys())
        image_number = 0
        for swath, polarization in product(swaths, polarizations):
            image_number += 1
            burst_infos = self.grouped_burst_infos[swath][polarization]
            swath = Swath(burst_infos, self.safe_path, self.version, self.creation_time, image_number)
            swath.assemble()
            swath.write()
            self.swaths.append(swath)

        for blank_product in self.create_blank_products(image_number):
            blank_product.assemble()
            swath_name = Swath.get_swath_name(blank_product.burst_infos, self.safe_path, blank_product.image_number)
            product_name = self.safe_path / 'annotation' / f'{swath_name}.xml'
            blank_product.write(product_name)
            self.blank_products.append(blank_product)

    def add_preview_components(self, content_units: List, metadata_objects: List, data_objects: List) -> List:
        """Add the preview components to unit lists.

        Args:
            content_units: A list of content units
            metadata_objects: A list of metadata objects
            data_objects: A list of data objects

        Returns:
            The updated content_units, metadata_objects, and data_objects lists
        """
        overlay_repid = 's1Level1MapOverlaySchema'
        preview_repid = 's1Level1ProductPreviewSchema'
        quicklook_repid = 's1Level1QuicklookSchema'
        overlay_content_unit = create_content_unit('mapoverlay', 'Metadata Unit', overlay_repid)
        preview_content_unit = create_content_unit('productpreview', 'Metadata Unit', preview_repid)
        quicklook_content_unit = create_content_unit('quicklook', 'Measurement Data Unit', quicklook_repid)
        content_units += [overlay_content_unit, preview_content_unit, quicklook_content_unit]

        metadata_objects += [create_metadata_object('mapoverlay'), create_metadata_object('productpreview')]

        # TOOD: add quciklook data object someday
        overlay_data_object = create_data_object(
            'mapoverlay',
            './preview/map-overlay.kml',
            overlay_repid,
            'text/xml',
            self.kml.size_bytes,
            self.kml.md5,
        )
        preview_data_object = create_data_object(
            'productpreview',
            './preview/product-preview.html',
            preview_repid,
            'text/html',
            self.preview.size_bytes,
            self.preview.md5,
        )
        data_objects += [overlay_data_object, preview_data_object]

        return content_units, metadata_objects, data_objects

    def compile_manifest_components(self) -> Tuple[List, List, List]:
        """Compile the manifest components for all files within the SAFE file.

        Returns:
            A list of content units, metadata objects, and data objects for the manifest file
        """
        content_units = []
        metadata_objects = []
        data_objects = []
        for swath in self.swaths:
            for annotation in swath.annotations:
                content_unit, metadata_object, date_object = annotation.create_manifest_components()
                content_units.append(content_unit)
                metadata_objects.append(metadata_object)
                data_objects.append(date_object)
            measurement_content, measurement_data = swath.measurement.create_manifest_components()
            content_units.append(measurement_content)
            data_objects.append(measurement_data)

        for blank_product in self.blank_products:
            content_unit, metadata_object, date_object = blank_product.create_manifest_components()
            content_units.append(content_unit)
            metadata_objects.append(metadata_object)
            data_objects.append(date_object)

        content_units, metadata_objects, data_objects = self.add_preview_components(
            content_units, metadata_objects, data_objects
        )
        return content_units, metadata_objects, data_objects

    def create_manifest(self) -> None:
        """Create the manifest.safe file for the SAFE file."""
        manifest_name = self.safe_path / 'manifest.safe'
        content_units, metadata_objects, data_objects = self.compile_manifest_components()
        template_manifest = get_subxml_from_metadata(self.burst_infos[0].metadata_path, 'manifest')
        manifest = Manifest(content_units, metadata_objects, data_objects, self.get_bbox(), template_manifest)
        manifest.assemble()
        manifest.write(manifest_name)
        self.manifest = manifest

    def create_preview(self):
        """Create the support files for the SAFE file."""
        kml = Kml(self.get_bbox())
        kml.assemble()
        kml.write(self.safe_path / 'preview' / 'map-overlay.kml')
        self.kml = kml

        product_names = [s.product_name.name for s in self.swaths]
        calibration_names = [s.noise_name.name for s in self.swaths] + [s.calibration_name.name for s in self.swaths]
        measurement_names = [s.measurement_name.name for s in self.swaths]
        rfi_names = [s.rfi_name.name for s in self.swaths if s.has_rfi]
        preview = Preview(self.name, product_names, calibration_names, measurement_names, rfi_names)
        preview.assemble()
        preview.write(self.safe_path / 'preview' / 'product-preview.html')
        self.preview = preview

    def update_product_identifier(self) -> None:
        """Update the product identifier using the CRC of the manifest file."""
        new_new = self.get_name(unique_id=self.manifest.crc)
        new_path = self.work_dir / new_new
        if new_path.exists():
            shutil.rmtree(new_path)
        shutil.move(self.safe_path, new_path)

        self.name = new_new
        self.safe_path = new_path

        self.kml.update_path(self.safe_path)
        self.preview.update_path(self.safe_path)
        for swath in self.swaths:
            swath.update_paths(self.safe_path)

    def create_safe(self) -> Path:
        """Create the SAFE file."""
        self.create_dir_structure()
        self.create_safe_components()
        self.create_preview()
        self.create_manifest()
        self.update_product_identifier()
        return self.safe_path

    def cleanup(self) -> None:
        """Remove unneeded files after SAFE creation"""
        to_delete = [burst_info.data_path for burst_info in self.burst_infos]
        to_delete += [burst_info.metadata_path for burst_info in self.burst_infos]
        to_delete = drop_duplicates(to_delete)
        for file in to_delete:
            file.unlink()



import json
import warnings
from argparse import Namespace
from binascii import crc_hqx
from collections.abc import Iterable
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, List, Optional

import asf_search
import lxml.etree as ET
from asf_search.Products.S1BurstProduct import S1BurstProduct
from osgeo import gdal, ogr, osr
from shapely.geometry import box, shape


gdal.UseExceptions()
warnings.filterwarnings('ignore')


@dataclass
class BurstInfo:
    """Dataclass for storing burst information."""

    granule: str
    slc_granule: str
    swath: str
    polarization: str
    burst_id: int
    burst_index: int
    direction: str
    absolute_orbit: int
    relative_orbit: int
    date: datetime
    data_url: Path
    data_path: Path
    metadata_url: Path
    metadata_path: Path
    start_utc: datetime = None
    stop_utc: datetime = None
    length: int = None
    width: int = None

    def add_shape_info(self):
        """Add shape information to the BurstInfo object."""
        annotation = get_subxml_from_metadata(self.metadata_path, 'product', self.swath, self.polarization)
        self.length = int(annotation.find('swathTiming/linesPerBurst').text)
        self.width = int(annotation.find('swathTiming/samplesPerBurst').text)

    def add_start_stop_utc(self):
        """Add start and stop UTC to burst info.
        There is spatial overlap between bursts, so burst start/stop times will overlap as well.
        """
        annotation = get_subxml_from_metadata(self.metadata_path, 'product', self.swath, self.polarization)
        start_utcs = [datetime.fromisoformat(x.find('azimuthTime').text) for x in annotation.findall('.//burst')]
        self.start_utc = start_utcs[self.burst_index]

        azimuth_time_interval = float(annotation.find('.//azimuthTimeInterval').text)
        burst_time_interval = timedelta(seconds=(self.length - 1) * azimuth_time_interval)
        self.stop_utc = self.start_utc + burst_time_interval


def create_burst_info(product: S1BurstProduct, work_dir: Path) -> BurstInfo:
    """Create a BurstInfo object given a granule.

    Args:
        product: A S1BurstProduct object
        work_dir: The directory to save the data to.
    """
    slc_granule = product.umm['InputGranules'][0].split('-')[0]

    burst_granule = product.properties['fileID']
    direction = product.properties['flightDirection'].upper()
    polarization = product.properties['polarization'].upper()
    absolute_orbit = int(product.properties['orbit'])
    relative_orbit = int(product.properties['pathNumber'])
    data_url = product.properties['url']
    metadata_url = product.properties['additionalUrls'][0]

    swath = product.properties['burst']['subswath'].upper()
    burst_id = int(product.properties['burst']['relativeBurstID'])
    burst_index = int(product.properties['burst']['burstIndex'])

    date_format = '%Y%m%dT%H%M%S'
    burst_time_str = burst_granule.split('_')[3]
    burst_time = datetime.strptime(burst_time_str, date_format)
    data_path = work_dir / f'{burst_granule}.tiff'
    metadata_path = work_dir / f'{slc_granule}_{polarization}.xml'

    burst_info = BurstInfo(
        burst_granule,
        slc_granule,
        swath,
        polarization,
        burst_id,
        burst_index,
        direction,
        absolute_orbit,
        relative_orbit,
        burst_time,
        data_url,
        data_path,
        metadata_url,
        metadata_path,
    )
    return burst_info


def get_burst_infos(products: Iterable[S1BurstProduct], work_dir: Path) -> List[BurstInfo]:
    """Get burst information from ASF Search.

    Args:
        products: A list of S1BurstProduct objects.
        save_dir: The directory to save the data to.

    Returns:
        A list of BurstInfo objects.
    """
    work_dir = optional_wd(work_dir)
    burst_info_list = []
    for product in products:
        burst_info = create_burst_info(product, work_dir)
        burst_info_list.append(burst_info)

    return burst_info_list


def sort_burst_infos(burst_info_list: List[BurstInfo]) -> Dict:
    """Sort BurstInfo objects by swath and polarization.

    Args:
        burst_info_list: List of BurstInfo objects.

    Returns:
        Dictionary of sorted BurstInfo objects. First key is swath, second key is polarization.
    """
    burst_infos = {}
    for burst_info in burst_info_list:
        if burst_info.swath not in burst_infos:
            burst_infos[burst_info.swath] = {}

        if burst_info.polarization not in burst_infos[burst_info.swath]:
            burst_infos[burst_info.swath][burst_info.polarization] = []

        burst_infos[burst_info.swath][burst_info.polarization].append(burst_info)

    swaths = list(burst_infos.keys())
    polarizations = list(burst_infos[swaths[0]].keys())
    for swath, polarization in zip(swaths, polarizations):
        burst_infos[swath][polarization] = sorted(burst_infos[swath][polarization], key=lambda x: x.burst_id)

    return burst_infos


def optional_wd(wd: Optional[str] = None) -> Path:
    """Return the working directory as a Path object

    Args:
        wd: Optional working directory as a Path or string

    Returns:
        Path to your input working directory or the current working directory.
    """
    if wd is None:
        wd = Path.cwd()
    return Path(wd).resolve()


def calculate_crc16(file_path: Path) -> str:
    """Calculate the CRC16 checksum for a file.

    Args:
        file_path: Path to file to calculate checksum for

    Returns:
        CRC16 checksum as a hexadecimal string
    """
    with open(file_path, 'rb') as f:
        data = f.read()

    crc = f'{crc_hqx(data, 0xffff):04X}'
    return crc


def get_subxml_from_metadata(
    metadata_path: Path, xml_type: str, subswath: str = None, polarization: str = None
) -> ET.Element:
    """Extract child xml info from ASF combined metadata file.

    Args:
        metadata_path: Path to metadata file
        xml_typ: Desired type of metadata to obtain (product, noise, calibration, or rfi)
        subswath: Desired subswath to obtain data for
        polarization: Desired polarization to obtain data for

    Returns:
        lxml Element for desired metadata
    """
    with open(metadata_path) as metadata_file:
        metadata = ET.parse(metadata_file).getroot()

    if xml_type == 'manifest':
        desired_metadata = metadata.find('manifest/{urn:ccsds:schema:xfdu:1}XFDU')
        return desired_metadata

    possible_types = ['product', 'noise', 'calibration', 'rfi']
    if xml_type not in possible_types:
        raise ValueError(f'Metadata type {xml_type} not one of {" ".join(possible_types)}')

    if subswath is None or polarization is None:
        raise ValueError('subswath and polarization must be provided for non-manifest files')

    correct_type = [x for x in metadata.find('metadata').iterchildren() if x.tag == xml_type]
    correct_swath = [x for x in correct_type if x.find('swath').text == subswath]
    correct_pol = [x for x in correct_swath if x.find('polarisation').text == polarization]

    if not correct_pol:
        desired_metadata = None
    else:
        desired_metadata = correct_pol[0].find('content')

    return desired_metadata


def download_url_with_retries(
    url: str, path: str, filename: str = None, session: asf_search.ASFSession = None, max_retries: int = 3
) -> None:
    """Download a file using asf_search.download_url with retries and backoff.

    Args:
        url: The URL to download
        path: The path to save the file to
        filename: The name of the file to save
        session: The ASF session to use
        max_retries: The maximum number of retries
    """
    n_retries = 0
    file_exists = False
    while n_retries < max_retries and not file_exists:
        asf_search.download_url(url, path, filename, session)

        n_retries += 1
        if Path(path, filename).exists():
            file_exists = True

    if not file_exists:
        raise ValueError(f'Failed to download {filename} after {max_retries} attempts.')


def flatten(list_of_lists: List[List]) -> List:
    """Flatten a list of lists."""
    return [item for sublist in list_of_lists for item in sublist]


def drop_duplicates(input_list: List) -> List:
    """Drop duplicates from a list, while preserving order."""
    return list(dict.fromkeys(input_list))


def set_text(element: ET.Element, text: str) -> None:
    """Set the text of an element if it is not None.

    Args:
        element: The element to set the text of.
        text: The text to set the element to.
    """
    if not isinstance(text, str) and not isinstance(text, int):
        raise ValueError('Text must be a string or an integer.')

    element.text = str(text)


def vector_to_shapely_latlon_polygon(vector_file_path):
    dataset = ogr.Open(vector_file_path)

    if dataset is None:
        raise ValueError(f'Could not open file: {vector_file_path}')

    layer = dataset.GetLayer()

    feature_count = layer.GetFeatureCount()
    if feature_count != 1:
        raise ValueError(f'File contains {feature_count} features, but exactly one is required.')

    feature = layer.GetFeature(0)
    geom = feature.GetGeometryRef()
    if geom.GetGeometryType() != ogr.wkbPolygon:
        raise ValueError('The feature is not a polygon.')

    source_srs = layer.GetSpatialRef()
    if int(source_srs.GetAuthorityCode(None)) != 4326:
        target_srs = osr.SpatialReference()
        target_srs.ImportFromEPSG(4326)
        transform = osr.CoordinateTransformation(source_srs, target_srs)
        geom.Transform(transform)

    polygon = shape(json.loads(geom.ExportToJson()))
    dataset = None

    return polygon


def reparse_args(args: Namespace, tool: str) -> Namespace:
    """Parse the arguments for burst2safe and burst2stack CLIs.

    Args:
        args: The parsed argument namespace.
        tool: The tool to parse arguments for (burst2safe or burst2stack).

    Returns:
        The parsed argument namespace.
    """
    arg_dict = args.__dict__
    tool_keywords = {
        'burst2safe': ['orbit', 'extent'],
        'burst2stack': ['rel_orbit', 'start_date', 'end_date', 'extent'],
    }
    keywords = tool_keywords[tool]

    using_granule = len(arg_dict.get('granules', [])) > 0
    used_keyword = [arg_dict.get(x, None) is not None for x in keywords]
    using_keywords = any(used_keyword)

    if arg_dict.get('mode', None) is None:
        args.mode = 'IW'

    if using_granule and using_keywords:
        raise ValueError(f'Cannot provide both granules and any of {", ".join(keywords)} arguments.')

    if not using_granule and not all(used_keyword):
        raise ValueError(f'Must provide at least {", ".join(keywords)} arguments.')

    if tool == 'burst2stack':
        args.start_date = datetime.strptime(args.start_date, '%Y-%m-%d')
        args.end_date = datetime.strptime(args.end_date, '%Y-%m-%d')

    if using_keywords:
        if args.mode not in ['IW', 'EW']:
            raise ValueError('--mode must be either IW or EW.')
        if args.pols:
            args.pols = [pol.upper() for pol in args.pols]
        if args.swaths:
            args.swaths = [swath.upper() for swath in args.swaths]

        if args.extent:
            try:
                args.extent = box(*[float(x) for x in args.extent])
            except ValueError:
                args.extent = vector_to_shapely_latlon_polygon(args.extent[0])
            except ValueError:
                raise ValueError('--extent cannot be interpreted as a bounding box or geometry file.')

    return args



"""A package for converting ASF burst SLCs to the SAFE format"""

import warnings
from collections.abc import Iterable
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from itertools import product
from multiprocessing import cpu_count
from pathlib import Path
from typing import List, Optional

import asf_search
import numpy as np
from asf_search.Products.S1BurstProduct import S1BurstProduct
from shapely.geometry import Polygon

from burst2safe.auth import get_earthdata_credentials
from burst2safe.utils import BurstInfo, download_url_with_retries


warnings.filterwarnings('ignore')


def find_granules(granules: Iterable[str]) -> List[S1BurstProduct]:
    """Find granules by name using ASF Search.

    Args:
        granules: A list of granule names

    Returns:
        A list of S1BurstProduct objects
    """
    results = asf_search.search(product_list=granules)
    found_granules = [result.properties['fileID'] for result in results]
    missing_granules = list(set(granules) - set(found_granules))
    if missing_granules:
        granule_str = ', '.joins(missing_granules)
        raise ValueError(f'Failed to find granule(s) {granule_str}. Check search parameters on Vertex.')
    return list(results)


def find_stack_orbits(rel_orbit: int, extent: Polygon, start_date: datetime, end_date: datetime) -> List[int]:
    """Find all orbits in a stack using ASF Search.

    Args:
        rel_orbit: The relative orbit number of the stack
        start_date: The start date of the stack
        end_date: The end date of the stack

    Returns:
        List of absolute orbit numbers
    """
    dataset = asf_search.constants.DATASET.SLC_BURST
    search_results = asf_search.geo_search(
        dataset=dataset,
        relativeOrbit=rel_orbit,
        intersectsWith=extent.centroid.wkt,
        start=start_date.strftime('%Y-%m-%d'),
        end=end_date.strftime('%Y-%m-%d'),
    )
    absolute_orbits = list(set([int(result.properties['orbit']) for result in search_results]))
    return absolute_orbits


def add_surrounding_bursts(bursts: List[S1BurstProduct], min_bursts: int) -> List[S1BurstProduct]:
    """Add bursts to the list to ensure each swath has at least `min_bursts` bursts.
    All bursts must be from the same absolute orbit, swath, and polarization.

    Args:
        bursts: A list of S1BurstProduct objects
        min_bursts: The minimum number of bursts

    Returns:
        An extended list of S1BurstProduct objects
    """
    ids = [int(burst.properties['burst']['relativeBurstID']) for burst in bursts]
    relative_orbit, _, swath = bursts[0].properties['burst']['fullBurstID'].split('_')
    polarization = bursts[0].properties['polarization']
    absolute_orbit = int(bursts[0].properties['orbit'])

    min_id, max_id = min(ids), max(ids)
    extra = np.floor((min_bursts - (max_id - min_id + 1)) / 2).astype(int)
    min_id -= extra
    max_id += extra
    if max_id - min_id + 1 != min_bursts:
        max_id += 1

    full_burst_ids = [f'{relative_orbit}_{id:06}_{swath}' for id in range(min_id, max_id + 1)]
    search_results = asf_search.search(
        dataset=asf_search.constants.DATASET.SLC_BURST,
        absoluteOrbit=absolute_orbit,
        polarization=polarization,
        fullBurstID=full_burst_ids,
    )
    return search_results


def find_swath_pol_group(
    search_results: List[S1BurstProduct], pol: str, swath: Optional[str], min_bursts: int
) -> List[S1BurstProduct]:
    """Find a group of bursts with the same polarization and swath.
    Add surrounding bursts if the group is too small.

    Args:
        search_results: A list of S1BurstProduct objects
        pol: The polarization to search for
        swath: The swath to search for
        min_bursts: The minimum number of bursts per swath

    Returns:
        An updated list of S1BurstProduct objects
    """
    if swath:
        search_results = [result for result in search_results if result.properties['burst']['subswath'] == swath]
    search_results = [result for result in search_results if result.properties['polarization'] == pol]

    params = [f'polarization {pol}']
    if swath:
        params.append(f'swath {swath}')
    params = ', '.join(params)

    if not search_results:
        raise ValueError(f'No bursts found for {params}. Check search parameters on Vertex.')

    if len(search_results) < min_bursts:
        search_results = add_surrounding_bursts(search_results, min_bursts)

    if len(search_results) < min_bursts:
        raise ValueError(f'Less than {min_bursts} bursts found for {params}. Check search parameters on Vertex.')

    return search_results


def find_group(
    orbit: int,
    footprint: Polygon,
    polarizations: Optional[Iterable] = None,
    swaths: Optional[Iterable] = None,
    mode: str = 'IW',
    min_bursts: int = 1,
) -> List[S1BurstProduct]:
    """Find burst groups using ASF Search.

    Args:
        orbit: The absolute orbit number of the bursts
        footprint: The bounding box of the bursts
        polarizations: List of polarizations to include (default: VV)
        swaths: List of swaths to include (default: all)
        mode: The collection mode to use (IW or EW) (default: IW)
        min_bursts: The minimum number of bursts per swath (default: 1)

    Returns:
        A list of S1BurstProduct objects
    """
    if polarizations is None:
        polarizations = ['VV']
    bad_pols = set(polarizations) - set(['VV', 'VH', 'HV', 'HH'])
    if bad_pols:
        raise ValueError(f'Invalid polarizations: {" ".join(bad_pols)}')

    if mode not in ['IW', 'EW']:
        raise ValueError('Invalid mode: must be IW or EW')
    elif mode == 'IW':
        valid_swaths = ['IW1', 'IW2', 'IW3']
    elif mode == 'EW':
        valid_swaths = ['EW1', 'EW2', 'EW3', 'EW4', 'EW5']

    if swaths is None:
        swaths = [None]
    else:
        bad_swaths = set(swaths) - set(valid_swaths)
        if bad_swaths:
            raise ValueError(f'Invalid swaths: {" ".join(bad_swaths)}')

    dataset = asf_search.constants.DATASET.SLC_BURST
    search_results = asf_search.geo_search(
        dataset=dataset, absoluteOrbit=orbit, intersectsWith=footprint.wkt, beamMode=mode
    )
    final_results = []
    for pol, swath in product(polarizations, swaths):
        sub_results = find_swath_pol_group(search_results, pol, swath, min_bursts)
        final_results.extend(sub_results)
    return final_results


def find_bursts(
    granules: Optional[Iterable[str]] = None,
    orbit: Optional[int] = None,
    footprint: Optional[Polygon] = None,
    polarizations: Optional[Iterable[str]] = None,
    swaths: Optional[Iterable[str]] = None,
    mode: str = 'IW',
    min_bursts: int = 1,
) -> Path:
    """Find bursts using ASF Search.

    Args:
        granules: A list of burst granules to convert to SAFE
        orbit: The absolute orbit number of the bursts
        footprint: The geographic extent of the bursts
        polarizations: List of polarizations to include
        swaths: List of swaths to include
        mode: The collection mode to use (IW or EW) (default: IW)
        min_bursts: The minimum number of bursts per swath (default: 1)
    """
    if granules:
        print('Using granule search...')
        results = find_granules(granules)
    elif orbit and footprint:
        print('Using burst group search...')
        results = find_group(orbit, footprint, polarizations, swaths, mode, min_bursts)
    else:
        raise ValueError(
            'You must provide either a list of granules or minimum set of group parameters (orbit, and footprint).'
        )
    return results


def download_bursts(burst_infos: Iterable[BurstInfo]) -> None:
    """Download the burst data and metadata files using multiple workers.

    Args:
        burst_infos: A list of BurstInfo objects
    """
    downloads = {}
    for burst_info in burst_infos:
        downloads[burst_info.data_path] = burst_info.data_url
        downloads[burst_info.metadata_path] = burst_info.metadata_url
    download_info = [(value, key.parent, key.name) for key, value in downloads.items()]
    urls, dirs, names = zip(*download_info)

    username, password = get_earthdata_credentials()
    session = asf_search.ASFSession().auth_with_creds(username, password)
    n_workers = min(len(urls), max(cpu_count() - 2, 1))
    if n_workers == 1:
        for url, dir, name in zip(urls, dirs, names):
            download_url_with_retries(url, dir, name, session)
    else:
        with ProcessPoolExecutor(max_workers=n_workers) as executor:
            executor.map(download_url_with_retries, urls, dirs, names, [session] * len(urls))



import netrc
import os
from pathlib import Path
from platform import system
from typing import Tuple


EARTHDATA_HOST = 'urs.earthdata.nasa.gov'


def get_netrc() -> Path:
    """Get the location of the netrc file.

    Returns:
        Path to the netrc file
    """
    netrc_name = '_netrc' if system().lower() == 'windows' else '.netrc'
    netrc_file = Path.home() / netrc_name
    return netrc_file


def find_creds_in_env(username_name, password_name) -> Tuple[str, str]:
    """Find credentials for a service in the environment.

    Args:
        username_name: Name of the environment variable for the username
        password_name: Name of the environment variable for the password

    Returns:
        Tuple of the username and password found in the environment
    """
    if username_name in os.environ and password_name in os.environ:
        username = os.environ[username_name]
        password = os.environ[password_name]
        return username, password

    return None, None


def find_creds_in_netrc(service) -> Tuple[str, str]:
    """Find credentials for a service in the netrc file.

    Args:
        service: Service to find credentials for

    Returns:
        Tuple of the username and password found in the netrc file
    """
    netrc_file = get_netrc()
    if netrc_file.exists():
        netrc_credentials = netrc.netrc(netrc_file)
        if service in netrc_credentials.hosts:
            username = netrc_credentials.hosts[service][0]
            password = netrc_credentials.hosts[service][2]
            return username, password

    return None, None


def get_earthdata_credentials() -> Tuple[str, str]:
    """Get NASA EarthData credentials from the environment or netrc file.

    Returns:
        Tuple of the NASA EarthData username and password
    """
    username, password = find_creds_in_env('EARTHDATA_USERNAME', 'EARTHDATA_PASSWORD')
    if username and password:
        return username, password

    username, password = find_creds_in_netrc(EARTHDATA_HOST)
    if username and password:
        return username, password

    raise ValueError(
        'Please provide NASA EarthData credentials via the '
        'EARTHDATA_USERNAME and EARTHDATA_PASSWORD environment variables, or your netrc file.'
    )



from pathlib import Path

from burst2safe import auth


def test_get_netrc(monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(auth, 'system', lambda: 'Windows')
        assert auth.get_netrc() == Path.home() / '_netrc'

    with monkeypatch.context() as m:
        m.setattr(auth, 'system', lambda: 'Linux')
        assert auth.get_netrc() == Path.home() / '.netrc'


def test_find_creds_in_env(monkeypatch):
    with monkeypatch.context() as m:
        m.setenv('TEST_USERNAME', 'foo')
        m.setenv('TEST_PASSWORD', 'bar')
        assert auth.find_creds_in_env('TEST_USERNAME', 'TEST_PASSWORD') == ('foo', 'bar')

    with monkeypatch.context() as m:
        m.delenv('TEST_USERNAME', raising=False)
        m.delenv('TEST_PASSWORD', raising=False)
        assert auth.find_creds_in_env('TEST_USERNAME', 'TEST_PASSWORD') == (None, None)


def test_find_creds_in_netrc(tmp_path, monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(auth, 'get_netrc', lambda: tmp_path / '.netrc')
        (tmp_path / '.netrc').write_text('machine test login foo password bar')
        assert auth.find_creds_in_netrc('test') == ('foo', 'bar')

    with monkeypatch.context() as m:
        m.setattr(auth, 'get_netrc', lambda: tmp_path / '.netrc')
        (tmp_path / '.netrc').write_text('')
        assert auth.find_creds_in_netrc('test') == (None, None)



"""A tool for converting ASF burst SLCs to the SAFE format"""

from argparse import ArgumentParser
from collections.abc import Iterable
from pathlib import Path
from typing import Optional

from shapely.geometry import Polygon

from burst2safe import utils
from burst2safe.safe import Safe
from burst2safe.search import download_bursts, find_bursts


DESCRIPTION = """Convert a set of ASF burst SLCs to the ESA SAFE format.
You can either provide a list of burst granules, or define a burst group by
providing the absolute orbit number, extent, and polarizations arguments.
"""


def burst2safe(
    granules: Optional[Iterable[str]] = None,
    orbit: Optional[int] = None,
    extent: Optional[Polygon] = None,
    polarizations: Optional[Iterable[str]] = None,
    swaths: Optional[Iterable[str]] = None,
    mode: str = 'IW',
    min_bursts: int = 1,
    all_anns: bool = False,
    keep_files: bool = False,
    work_dir: Optional[Path] = None,
) -> Path:
    """Convert a set of burst granules to the ESA SAFE format.

    To be eligible for conversions, all burst granules must:
    - Have the same acquisition mode
    - Be from the same absolute orbit
    - Be contiguous in time and space
    - Have the same footprint for all included polarizations

    Args:
        granules: A list of burst granules to convert to SAFE
        orbit: The absolute orbit number of the bursts
        extent: The bounding box of the bursts
        polarizations: List of polarizations to include
        swaths: List of swaths to include
        mode: The collection mode to use (IW or EW)
        min_bursts: The minimum number of bursts per swath (default: 1)
        all_anns: Include product annotation files for all swaths, regardless of included bursts
        keep_files: Keep the intermediate files
        work_dir: The directory to create the SAFE in (default: current directory)
    """
    work_dir = utils.optional_wd(work_dir)

    products = find_bursts(granules, orbit, extent, polarizations, swaths, mode, min_bursts)
    burst_infos = utils.get_burst_infos(products, work_dir)
    print(f'Found {len(burst_infos)} burst(s).')

    print('Check burst group validity...')
    Safe.check_group_validity(burst_infos)

    print('Downloading data...')
    download_bursts(burst_infos)
    print('Download complete.')

    print('Creating SAFE...')
    [info.add_shape_info() for info in burst_infos]
    [info.add_start_stop_utc() for info in burst_infos]

    safe = Safe(burst_infos, all_anns, work_dir)
    safe_path = safe.create_safe()
    print('SAFE created!')

    if not keep_files:
        safe.cleanup()

    return safe_path


def main() -> None:
    parser = ArgumentParser(description=DESCRIPTION)
    parser.add_argument('granules', nargs='*', help='List of bursts to convert to SAFE')
    parser.add_argument('--orbit', type=int, help='Absolute orbit number of the bursts')
    parser.add_argument(
        '--extent',
        type=str,
        nargs='+',
        help='Bounds (W S E N in lat/lon) or geometry file describing spatial extent',
    )
    parser.add_argument('--pols', type=str, nargs='+', help='Polarizations to include (i.e., VV VH). Default: VV')
    parser.add_argument(
        '--swaths', type=str, nargs='+', help='Swaths to include (i.e., IW1 IW2 IW3). Defaults to all swaths.'
    )
    parser.add_argument('--mode', type=str, default='IW', help='Collection mode to use (IW or EW). Default: IW')
    parser.add_argument('--min-bursts', type=int, default=1, help='Minimum # of bursts per swath/polarization.')
    parser.add_argument(
        '--all-anns',
        action='store_true',
        default=False,
        help='Include product annotations files for all swaths, regardless of included bursts.',
    )
    parser.add_argument('--output-dir', type=str, default=None, help='Output directory to save to')
    parser.add_argument('--keep-files', action='store_true', default=False, help='Keep the intermediate files')

    args = utils.reparse_args(parser.parse_args(), tool='burst2safe')

    burst2safe(
        granules=args.granules,
        orbit=args.orbit,
        extent=args.extent,
        polarizations=args.pols,
        swaths=args.swaths,
        mode=args.mode,
        min_bursts=args.min_bursts,
        all_anns=args.all_anns,
        keep_files=args.keep_files,
        work_dir=args.output_dir,
    )



import filecmp
import shutil
import subprocess
from pathlib import Path

import pytest
from shapely import box

from burst2safe.burst2safe import burst2safe


CURRENT_BRANCH = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip().decode('utf-8')


def bit_for_bit(reference: Path, secondary: Path):
    filecmp.clear_cache()
    return filecmp.cmp(reference, secondary, shallow=False)


@pytest.mark.golden()
@pytest.mark.dependency()
def test_burst2safe_iw():
    work_path = Path.cwd() / CURRENT_BRANCH / 'iw'
    work_path.mkdir(parents=True, exist_ok=True)
    burst2safe(
        granules=[],
        orbit=51936,
        extent=box(*[-117.3, 35.5, -117.2, 35.6]),
        polarizations=['VV', 'VH'],
        keep_files=False,
        work_dir=work_path,
    )


@pytest.mark.golden()
@pytest.mark.dependency()
def test_burst2safe_ew():
    work_path = Path.cwd() / CURRENT_BRANCH / 'ew'
    work_path.mkdir(parents=True, exist_ok=True)
    burst2safe(
        granules=[],
        orbit=54631,
        extent=box(*[-53.6, 66.6, -53.3, 66.8]),
        polarizations=['HH', 'HV'],
        mode='EW',
        keep_files=False,
        work_dir=work_path,
    )


@pytest.mark.golden()
@pytest.mark.dependency(depends=['test_burst2safe_iw', 'test_burst2safe_ew'])
def test_make_archive():
    work_path = Path.cwd() / CURRENT_BRANCH
    shutil.make_archive(work_path, 'tar', work_path)


@pytest.mark.golden()
@pytest.mark.parametrize('mode', ['iw', 'ew'])
def test_golden_compare(mode):
    main_branch = 'main'
    develop_branch = 'develop'

    for tar_name in [f'{main_branch}.tar', f'{develop_branch}.tar']:
        tar_path = Path.cwd() / tar_name
        extract_dir = tar_path.with_suffix('')
        if not extract_dir.exists():
            extract_dir.mkdir(exist_ok=True)
            shutil.unpack_archive(tar_path, extract_dir)

    main_safe = list((Path.cwd() / main_branch / mode).glob('*.SAFE'))[0]
    develop_safe = list((Path.cwd() / develop_branch / mode).glob('*.SAFE'))[0]
    assert main_safe.name == develop_safe.name

    main_files = sorted([x.resolve() for x in main_safe.rglob('*')])
    develop_files = sorted([x.resolve() for x in develop_safe.rglob('*')])

    main_set = set([f.relative_to(main_safe) for f in main_files])
    develop_set = set([f.relative_to(develop_safe) for f in develop_files])

    only_in_main = [str(x) for x in main_set - develop_set]
    assert not only_in_main

    only_in_develop = [str(x) for x in develop_set - main_set]
    assert not only_in_develop

    differing_files = []
    for main_file, develop_file in zip(main_files, develop_files):
        if main_file.is_dir():
            continue

        if not bit_for_bit(main_file, develop_file):
            differing_files.append(develop_file.relative_to(develop_safe))

    differing_files = [str(x) for x in differing_files]
    assert not differing_files



import hashlib
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path
from typing import Tuple

import numpy as np
from osgeo import gdal, osr
from tifffile import TiffFile

from burst2safe.base import create_content_unit, create_data_object
from burst2safe.product import GeoPoint
from burst2safe.utils import BurstInfo


class Measurement:
    """Class representing a measurement GeoTIFF."""

    def __init__(
        self,
        burst_infos: Iterable[BurstInfo],
        gcps: Iterable[GeoPoint],
        creation_time: datetime,
        ipf_version: str,
        image_number: int,
    ):
        """Initialize a Measurement object.

        Args:
            burst_infos: A list of BurstInfo objects
            gcps: A list of GeoPoint objects
            creation_time: The creation time of the measurement
            ipf_version: The IPF version of the measurement data
            image_number: The image number of the measurement
        """
        self.burst_infos = burst_infos
        self.gcps = gcps
        self.creation_time = creation_time
        self.version = ipf_version
        self.image_number = image_number

        self.swath = self.burst_infos[0].swath
        self.s1_platform = self.burst_infos[0].slc_granule[2].upper()

        burst_lengths = sorted(list(set([info.length for info in burst_infos])))
        if len(burst_lengths) != 1:
            raise ValueError(f'All burst are not the same length. Found {" ".join([str(x) for x in burst_lengths])}')
        self.burst_length = burst_lengths[0]
        self.total_length = self.burst_length * len(self.burst_infos)

        # TODO: sometimes bursts from different SLCs have different widths. Is this an issue?
        self.total_width = max([info.width for info in burst_infos])

        self.data_mean = None
        self.data_std = None
        self.size_bytes = None
        self.md5 = None
        self.byte_offsets = []

    def get_data(self, band: int = 1) -> np.ndarray:
        """Get the data for the measurement from ASF burst GeoTIFFs.

        Args:
            band: The GeoTIFF band to read

        Returns:
            The data from burst GeoTIFFs as a numpy array
        """
        data = np.zeros((self.total_length, self.total_width), dtype=np.complex64)
        for i, burst_info in enumerate(self.burst_infos):
            ds = gdal.Open(str(burst_info.data_path))
            burst_slice = np.s_[i * self.burst_length : (i + 1) * self.burst_length, 0 : burst_info.width]
            data[burst_slice] = ds.GetRasterBand(band).ReadAsArray()
            ds = None
        return data

    def get_burst_byte_offsets(self):
        """Get the byte offsets of each burst in the measurement GeoTIFF."""
        with TiffFile(self.path) as tif:
            if len(tif.pages) != 1:
                raise ValueError('Byte offset calculation only valid for GeoTIFFs with one band.')
            page = tif.pages[0]

            if page.compression._name_ != 'NONE':
                raise ValueError('Byte offset calculation only valid for uncompressed GeoTIFFs.')

            if page.chunks != (1, page.imagewidth):
                raise ValueError('Byte offset calculation only valid for GeoTIFFs with one line per block.')

            offsets = page.dataoffsets

        byte_offsets = [offsets[self.burst_length * i] for i in range(len(self.burst_infos))]
        return byte_offsets

    def add_metadata(self, dataset: gdal.Dataset):
        """Add metadata to an existing GDAL dataset.

        Args:
            dataset: The GDAL dataset to add metadata to
        """
        gdal_gcps = [gdal.GCP(gcp.x, gcp.y, gcp.z, gcp.pixel, gcp.line) for gcp in self.gcps]
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(4326)
        dataset.SetGCPs(gdal_gcps, srs.ExportToWkt())

        dataset.SetMetadataItem('TIFFTAG_DATETIME', datetime.strftime(self.creation_time, '%Y:%m:%d %H:%M:%S'))
        dataset.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION', f'Sentinel-1{self.s1_platform} IW SLC L1')
        dataset.SetMetadataItem('TIFFTAG_SOFTWARE', f'Sentinel-1 IPF {self.version}')

    def create_geotiff(self, out_path: Path, update_info=True):
        """Create a GeoTIFF of SLC data from the constituent burst SLC GeoTIFFs.
        Optionally update Measurment metadata.

        Args:
            out_path: The path to write the SLC GeoTIFF to
            update_info: Whether to update the Measurement metadata
        """
        gtiff = gdal.GetDriverByName('GTiff')
        blank_ds = gtiff.Create(str(out_path), self.total_width, self.total_length, 1, gdal.GDT_CInt16)
        blank_band = blank_ds.GetRasterBand(1)
        blank_band.SetNoDataValue(0)
        blank_band.FlushCache()
        blank_ds = None

        ds = gdal.Open(str(out_path), gdal.GA_Update)
        band = ds.GetRasterBand(1)
        data = self.get_data()
        band.WriteArray(data)
        self.add_metadata(ds)
        band.FlushCache()
        ds = None

        if update_info:
            self.path = out_path
            self.data_mean = np.mean(data[data != 0])
            self.data_std = np.std(data[data != 0])
            self.byte_offsets = self.get_burst_byte_offsets()

            with open(self.path, 'rb') as f:
                file_bytes = f.read()
                self.size_bytes = len(file_bytes)
                self.md5 = hashlib.md5(file_bytes).hexdigest()

    def create_manifest_components(self) -> Tuple:
        """Create the components of the SAFE manifest for the measurement file.

        Returns:
            A tuple of the content unit and data object for the measurement file
        """
        simple_name = self.path.with_suffix('').name.replace('-', '')
        rep_id = 's1Level1MeasurementSchema'
        unit_type = 'Measurement Data Unit'
        mime_type = 'application/octet-stream'

        safe_index = [i for i, x in enumerate(self.path.parts) if 'SAFE' in x][-1]
        safe_path = Path(*self.path.parts[: safe_index + 1])
        relative_path = self.path.relative_to(safe_path)

        content_unit = create_content_unit(simple_name, rep_id, unit_type)
        data_object = create_data_object(simple_name, relative_path, rep_id, mime_type, self.size_bytes, self.md5)
        return content_unit, data_object

    def write(self, out_path):
        """Write the measurement GeoTIFF to a file."""
        self.create_geotiff(out_path)



"""Calculate the unique burst ID of a Sentinel-1 burst using ESA's convention.
Created by Ben Barton for ASF's Sentinel-1 Burst ingest utility.

Source:
https://github.com/asfadmin/rain/blob/a5cd0709d46f93de1060ddb3fc7673097a9a966d/code/util/burst.py
"""

from datetime import datetime, timedelta

import numpy as np
from dateutil import parser as dateparser


# These constants are from the Sentinel-1 Level 1 Detailed Algorithm Definition PDF
# MPC Nom: DI-MPC-IPFDPM, MPC Ref: MPC-0307, Issue/Revision: 2/4, Table 9-7
NOMINAL_ORBITAL_DURATION = timedelta(seconds=12 * 24 * 3600 / 175)
PREAMBLE_LENGTH_IW = timedelta(seconds=2.299849)
PREAMBLE_LENGTH_EW = timedelta(seconds=2.299970)
BEAM_CYCLE_TIME_IW = timedelta(seconds=2.758273)
BEAM_CYCLE_TIME_EW = timedelta(seconds=3.038376)
MODE_TIMING = {
    'EW': (
        PREAMBLE_LENGTH_EW,
        BEAM_CYCLE_TIME_EW,
    ),
    'IW': (
        PREAMBLE_LENGTH_IW,
        BEAM_CYCLE_TIME_IW,
    ),
}


class InvalidModeNameError(Exception):
    pass


def _get_mode_timing(mode_name: str) -> tuple:
    """Get the timing parameters for a given mode.

    Args:
        mode_name: Mode name (EW,IW) or subswath name (EW2 etc)

    Returns:
        PREAMBLE_LENGTH, BEAM_CYCLE_TIME for passed mode
    """
    try:
        return MODE_TIMING[mode_name[:2].upper()]
    except KeyError as e:
        raise InvalidModeNameError(e)


def _calc_start_subsw1(start_offsets: list, swath_num: int, sensing_time: datetime) -> datetime:
    """Calculates the start time of the #1 subswath given any subswath in a burst.

    Args:
        start_offsets: list of start offsets, times since IW1
        swath_num: swath number.
        sensing_time: datetime of the sensing time of the first input line of this burst [UTC]

    Returns:
        datetime of the start of the first subswath
    """
    offset = start_offsets[swath_num - 1]
    start_subsw1 = sensing_time + timedelta(seconds=offset)
    return start_subsw1


def _calc_esa_burstid(
    time_since_anx: timedelta, orbit_number_start: int, preamble_len: timedelta, beam_cycle_time: timedelta
) -> int:
    """Calculates the ESA Burst ID given the given parameters.
    Extracted from the opera function for the purposes of reuse IW & EW

    Args:
        time_since_anx: time since ascending node crossing
        orbit_number_start: Relative orbit number at the start of the acquisition
        preamble_len: Preamble length
        beam_cycle_time: Beam cycle time

    Returns:
        ESA Burst ID
    """
    # Eq. 9-89: ∆tb = tb − t_anx + (r - 1)T_orb
    # tb: mid-burst sensing time (sensing_time)
    # t_anx: ascending node time (ascending_node_dt)
    # r: relative orbit number   (relative_orbit_start)
    dt_b = time_since_anx + (orbit_number_start - 1) * NOMINAL_ORBITAL_DURATION

    # Eq. 9-91 :   1 + floor((∆tb − T_pre) / T_beam )
    esa_burst_id = 1 + int(np.floor((dt_b - preamble_len) / beam_cycle_time))
    return esa_burst_id


def calculate_burstid(
    sensing_time_iso: str,
    anx_time_iso: str,
    orbit_number_start: int,
    orbit_number_stop: int,
    subswath: str,
) -> tuple:
    """Calculate the unique burst ID of a burst. And return the
    correct orbit number.

    Accounts for equator crossing frames, where the current track number of
    a burst may change mid-frame. Uses the ESA convention defined in the
    Sentinel-1 Level 1 Detailed Algorithm Definition.

    Args:
        sensing_time_iso: Iso date of sensing time of the first input line of this burst [UTC]
        The XML tag is sensingTime in the annotation file.
        anx_time_iso: Time of the ascending node prior to the start of the scene.
        orbit_number_start: Relative orbit number at the start of the acquisition, from 1-175.
        orbit_number_stop: Relative orbit number at the end of the acquisition.
        subswath: Name of the subswath of the burst (not case sensitive).

    Returns:
        burst ID, orbit number

    Notes:
    The `orbit_number_start` and `orbit_number_stop` parameters are used to determine if the
    scene crosses the equator. They are the same if the frame does not cross
    the equator.

    References:
    ESA Sentinel-1 Level 1 Detailed Algorithm Definition
    https://sentinels.copernicus.eu/documents/247904/1877131/S1-TN-MDA-52-7445_Sentinel-1+Level+1+Detailed+Algorithm+Definition_v2-4.pdf/83624863-6429-cfb8-2371-5c5ca82907b8

    Stolen from https://github.com/opera-adt/s1-reader/blob/main/src/s1reader/s1_burst_id.py
    """
    sensing_dt = dateparser.parse(sensing_time_iso)
    asc_node_dt = dateparser.parse(anx_time_iso)
    mode = subswath[0:2].upper()
    swath_num = int(subswath[-1])
    pream, beamtime = _get_mode_timing(mode)

    # Since we only have access to the current subswath, we need to use the
    # burst-to-burst times to figure out
    #   1. if IW1 crossed the equator, and
    #   2. The mid-burst sensing time for IW2
    # IW1 -> IW2 takes ~0.83220 seconds
    # IW2 -> IW3 takes ~1.07803 seconds
    # IW3 -> IW1 takes ~0.84803 seconds
    # EW1 -> EW2 ~0.6826778977095476
    # EW2 -> EW3 ~0.558731729405626
    # EW3 -> EW4 ~0.6123389197595133
    # EW4 -> EW5 ~0.5653802400407812
    # EW5 -> EW1 ~0.619247213084539

    burst_times = {'IW': [0.832, 1.078, 0.848], 'EW': [0.683, 0.559, 0.612, 0.565, 0.619]}

    if mode == 'IW':
        iw1_start_offsets = [
            0,
            -burst_times[mode][0],
            -burst_times[mode][0] - burst_times[mode][1],
        ]

        start_iw1 = _calc_start_subsw1(iw1_start_offsets, swath_num, sensing_dt)

        # Middle of IW2 is the middle of the entire burst:
        start_iw1_to_mid_iw2 = burst_times[mode][0] + burst_times[mode][1] / 2
        mid_iw2 = start_iw1 + timedelta(seconds=start_iw1_to_mid_iw2)

        has_anx_crossing = (orbit_number_stop == orbit_number_start + 1) or (
            orbit_number_stop == 1 and orbit_number_start == 175
        )

        time_since_anx_iw1 = start_iw1 - asc_node_dt
        time_since_anx = mid_iw2 - asc_node_dt

        if (time_since_anx_iw1 - NOMINAL_ORBITAL_DURATION).total_seconds() < 0:
            # Less than a full orbit has passed
            orbit_number = orbit_number_start
        else:
            orbit_number = orbit_number_stop
            # Additional check for scenes which have a given ascending node
            # that's more than 1 orbit in the past
            if not has_anx_crossing:
                time_since_anx = time_since_anx - NOMINAL_ORBITAL_DURATION

        esa_burst_id = _calc_esa_burstid(time_since_anx, orbit_number_start, pream, beamtime)

    elif mode == 'EW':
        # EWs are always high-latitude and don't cross equator. We can always
        # use start orbit.
        orbit_number = orbit_number_start

        ew1_start_offsets = [
            0,
            -burst_times[mode][0],
            -burst_times[mode][0] - burst_times[mode][1],
            -burst_times[mode][0] - burst_times[mode][1] - burst_times[mode][2],
            -burst_times[mode][0] - burst_times[mode][1] - burst_times[mode][2] - burst_times[mode][3],
        ]

        start_ew1 = _calc_start_subsw1(ew1_start_offsets, swath_num, sensing_dt)

        # Middle of EW3 is the middle of the entire burst:
        start_ew1_to_mid_ew3 = burst_times[mode][0] + burst_times[mode][1] + burst_times[mode][2] / 2
        mid_ew3 = start_ew1 + timedelta(seconds=start_ew1_to_mid_ew3)

        time_since_anx = mid_ew3 - asc_node_dt

        esa_burst_id = _calc_esa_burstid(time_since_anx, orbit_number_start, pream, beamtime)

    else:
        raise InvalidModeNameError('Invalid mode name: %s' % mode)

    return esa_burst_id, orbit_number



from collections.abc import Iterable
from copy import deepcopy

import lxml.etree as ET

from burst2safe.base import Annotation
from burst2safe.utils import BurstInfo


class Rfi(Annotation):
    """Class representing an radio frequency interference (RFI) XML.

    Note: RFI annotations only available for IPF version 3.40 onwards.
    """

    def __init__(self, burst_infos: Iterable[BurstInfo], ipf_version: str, image_number: int):
        """Create a calibration object.

        Args:
            burst_infos: List of BurstInfo objects.
            ipf_version: The IPF version of the annotation (i.e. 3.71).
            image_number: Image number.
        """
        super().__init__(burst_infos, 'rfi', ipf_version, image_number)
        self.rfi_mitigation_applied = None
        self.rfi_detection_from_noise_report_list = None
        self.rfi_burst_report_list = None

    def create_rfi_mitigation_applied(self):
        """Create the rifMitigationApplied element."""
        self.rfi_mitigation_applied = deepcopy(self.inputs[0].find('rfiMitigationApplied'))

    def create_rfi_detection_from_noise_report_list(self):
        """Create the rfiDetectionFromNoiseReportList element."""
        self.rfi_detection_from_noise_report_list = self.merge_lists('rfiDetectionFromNoiseReportList')

    def create_rfi_burst_report_list(self):
        """Create the rfiBurstReportList element."""
        self.rfi_burst_report_list = self.merge_lists('rfiBurstReportList')

    def assemble(self):
        """Assemble the RFI object components."""
        self.create_ads_header()
        self.create_rfi_mitigation_applied()
        self.create_rfi_detection_from_noise_report_list()
        self.create_rfi_burst_report_list()

        rfi = ET.Element('rfi')
        rfi.append(self.ads_header)
        rfi.append(self.rfi_mitigation_applied)
        rfi.append(self.rfi_detection_from_noise_report_list)
        rfi.append(self.rfi_burst_report_list)
        rfi_tree = ET.ElementTree(rfi)

        ET.indent(rfi_tree, space='  ')
        self.xml = rfi_tree



import lxml.etree as ET
import pytest

from burst2safe.burst2safe import burst2safe


@pytest.mark.integration()
def test_all_anns(tmp_path):
    safe_path = burst2safe(granules=['S1_088916_IW3_20240605T140826_VV_36F4-BURST'], all_anns=True, work_dir=tmp_path)
    products = list((safe_path / 'annotation').glob('*.xml'))
    assert len(products) == 3

    for swath in ['iw1', 'iw2']:
        prod = [p for p in products if swath in p.name][0]
        xml = ET.parse(prod).getroot()

        burst_list = xml.find('.//burstList')
        assert burst_list.attrib['count'] == '0'
        assert len(burst_list) == 0

        geolocation_grid = xml.find('.//geolocationGridPointList')
        assert geolocation_grid.attrib['count'] == '0'
        assert len(geolocation_grid) == 0

    prod_real = [p for p in products if 'iw3' in p.name][0]
    xml_real = ET.parse(prod_real).getroot()

    burst_list = xml_real.find('.//burstList')
    assert int(burst_list.attrib['count']) > 0
    assert len(burst_list) > 0

    geolocation_grid = xml_real.find('.//geolocationGridPointList')
    assert int(geolocation_grid.attrib['count']) > 0
    assert len(geolocation_grid) > 0



import hashlib
from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional

import lxml.etree as ET

from burst2safe.utils import BurstInfo, drop_duplicates, flatten, get_subxml_from_metadata, set_text


SCHEMA = '{urn:ccsds:schema:xfdu:1}'


class ListOfListElements:
    def __init__(
        self, inputs: List[ET.Element], start_line: Optional[int] = None, slc_lengths: Optional[List[int]] = None
    ):
        """Initialize the ListOfListElements object.

        Args:
            inputs: The list of elements to be processed.
            start_line: The starting line number of the first element.
            slc_lengths: The total line lengths of the SLCs corresponding to each element.
        """
        self.inputs = inputs
        self.start_line = start_line
        self.slc_lengths = slc_lengths

        self.name = self.inputs[0].tag
        elements = flatten([element.findall('*') for element in self.inputs])
        if len(elements) == 0:
            raise ValueError(f'No Sub-elements contained within {self.name}.')

        names = drop_duplicates([x.tag for x in elements])
        if len(names) != 1:
            raise ValueError('Elements must contain only one type of subelement.')
        self.subelement_name = names[0]

        if self.name == 'replicaInformationList':
            self.time_field = 'referenceReplica/azimuthTime'
        elif 'azimuthTime' in [x.tag for x in elements[0].iter()]:
            self.time_field = 'azimuthTime'
        elif 'time' in [x.tag for x in elements[0].iter()]:
            self.time_field = 'time'
        elif 'noiseSensingTime' in [x.tag for x in elements[0].iter()]:
            self.time_field = 'noiseSensingTime'
        else:
            raise ValueError('Time field not found in elements.')

        self.inputs = sorted(self.inputs, key=self.get_first_time)
        self.has_line = elements[0].find('line') is not None

    def get_first_time(self, element: ET.Element) -> datetime:
        """Get first time in List element

        Args:
            element: The element to get the time from.

        Returns:
            The first time in the element.
        """
        first_time = min([datetime.fromisoformat(sub.find(self.time_field).text) for sub in element])
        return first_time

    def get_unique_elements(self) -> List[ET.Element]:
        """Get the elements without duplicates. Adjust line number if present.

        Returns:
            The list of elements without duplicates.
        """
        list_of_element_lists = [item.findall('*') for item in self.inputs]

        last_time = datetime.fromisoformat(list_of_element_lists[0][-1].find(self.time_field).text)
        uniques = [deepcopy(element) for element in list_of_element_lists[0]]
        if self.has_line:
            previous_line_count = self.slc_lengths[0]

        for i, element_list in enumerate(list_of_element_lists[1:]):
            times = [datetime.fromisoformat(element.find(self.time_field).text) for element in element_list]
            keep_index = [index for index, time in enumerate(times) if time > last_time]
            to_keep = [deepcopy(element_list[index]) for index in keep_index]

            if self.has_line:
                new_lines = [int(elem.find('line').text) + previous_line_count for elem in to_keep]
                [set_text(elem.find('line'), line) for elem, line in zip(to_keep, new_lines)]
                previous_line_count += self.slc_lengths[i]

            last_time = max([times[index] for index in keep_index])
            uniques += to_keep

        return uniques

    @staticmethod
    def filter_by_line(element_list: List[ET.Element], line_bounds: tuple[float, float]) -> List[ET.Element]:
        """Filter elements by line number.

        Args:
            line_bounds: The bounds of the line numbers.

        Returns:
            A filtered element list.
        """
        new_list = []
        for elem in element_list:
            if line_bounds[0] <= int(elem.find('line').text) <= line_bounds[1]:
                new_list.append(deepcopy(elem))
        return new_list

    def update_line_numbers(self, elements: List[ET.Element]) -> None:
        """Update the line numbers of the elements.

        Args:
            elements: The list of elements to update.
        """
        for element in elements:
            standard_line = int(element.find('line').text)
            element.find('line').text = str(standard_line - self.start_line)

    def filter_by_time(
        self, elements: List[ET.Element], anx_bounds: tuple[float, float], buffer: timedelta
    ) -> List[ET.Element]:
        """Filter elements by time.

        Args:
            elements: The list of elements to filter.
            anx_bounds: The bounds of the ANX time.
            buffer: The buffer to add to the ANX bounds.

        Returns:
            A filtered element list.
        """
        min_anx_bound = anx_bounds[0] - buffer
        max_anx_bound = anx_bounds[1] + buffer
        filtered_elements = []
        for element in elements:
            azimuth_time = datetime.fromisoformat(element.find(self.time_field).text)
            if min_anx_bound < azimuth_time < max_anx_bound:
                filtered_elements.append(deepcopy(element))

        return filtered_elements

    def create_filtered_list(
        self,
        anx_bounds: Optional[tuple[float, float]],
        buffer: Optional[timedelta] = timedelta(seconds=3),
        line_bounds: Optional[tuple[float, float]] = None,
    ) -> ET.Element:
        """Filter elements by time/line. Adjust line number if present.

        Args:
            anx_bounds: The bounds of the ANX time.
            buffer: The buffer to add to the ANX bounds.
            line_bounds: The bounds of the line numbers.

        Returns:
            A filtered list element.
        """
        elements = self.get_unique_elements()
        filtered_elements = self.filter_by_time(elements, anx_bounds, buffer)

        if self.has_line:
            self.update_line_numbers(filtered_elements)

        if line_bounds is not None:
            if not self.has_line:
                raise ValueError('Line bounds cannot be applied to elements without line numbers.')
            filtered_elements = self.filter_by_line(filtered_elements, line_bounds)

        new_element = ET.Element(self.name)
        [new_element.append(element) for element in filtered_elements]
        new_element.set('count', str(len(filtered_elements)))
        return new_element


def create_content_unit(simple_name: str, unit_type: str, rep_id: str) -> ET.Element:
    """Create a content unit element for a manifest.safe file.

    Args:
        simple_name: The simple name of the content unit.
        unit_type: The type of the content unit.
        rep_id: The representation ID.

    Returns:
        The content unit element.
    """
    content_unit = ET.Element(f'{SCHEMA}contentUnit')
    content_unit.set('unitType', unit_type)
    content_unit.set('repID', rep_id)
    ET.SubElement(content_unit, 'dataObjectPointer', dataObjectID=simple_name)
    return content_unit


def create_metadata_object(simple_name: str) -> ET.Element:
    """Create a metadata object element for a manifest.safe file.

    Args:
        simple_name: The simple name of the metadata object.

    Returns:
        The metadata object element.
    """
    metadata_object = ET.Element('metadataObject')
    metadata_object.set('ID', f'{simple_name}Annotation')
    metadata_object.set('classification', 'DESCRIPTION')
    metadata_object.set('category', 'DMD')
    ET.SubElement(metadata_object, 'dataObjectPointer', dataObjectID=simple_name)
    return metadata_object


def create_data_object(
    simple_name: str, relative_path: Path, rep_id: str, mime_type: str, size_bytes: int, md5: str
) -> ET.Element:
    """Create a data object element for a manifest.safe file.

    Args:
        simple_name: The simple name of the data object.
        relative_path: The relative path to the data object.
        rep_id: The representation ID.
        mime_type: The MIME type of the data object.
        size_bytes: The size of the data object in bytes.
        md5: The MD5 checksum of the data object.

    Returns:
        The data object element.
    """
    data_object = ET.Element('dataObject')
    data_object.set('ID', simple_name)
    data_object.set('repID', rep_id)
    byte_stream = ET.SubElement(data_object, 'byteStream')
    byte_stream.set('mimeType', mime_type)
    byte_stream.set('size', str(size_bytes))
    file_location = ET.SubElement(byte_stream, 'fileLocation')
    file_location.set('locatorType', 'URL')
    file_location.set('href', f'./{relative_path}')
    checksum = ET.SubElement(byte_stream, 'checksum')
    checksum.set('checksumName', 'MD5')
    checksum.text = md5
    return data_object


class Annotation:
    def __init__(self, burst_infos: Iterable[BurstInfo], metadata_type: str, ipf_version: str, image_number: int):
        """Initialize the Annotation object.

        Args:
            burst_infos: The list of burst information objects.
            metadata_type: The type of metadata to create.
            ipf_version: The IPF version of the annotation (i.e. 3.71).
            image_number: The image number of the annotation.
        """
        self.burst_infos = burst_infos
        self.metadata_type = metadata_type
        self.image_number = image_number
        self.major_version, self.minor_version = [int(v) for v in ipf_version.split('.')]
        self.metadata_paths = drop_duplicates([x.metadata_path for x in burst_infos])
        self.swath, self.pol = burst_infos[0].swath, burst_infos[0].polarization
        self.start_line = burst_infos[0].burst_index * burst_infos[0].length
        self.total_lines = len(burst_infos) * burst_infos[0].length
        self.stop_line = self.start_line + self.total_lines
        self.min_anx = min([x.start_utc for x in burst_infos])
        self.max_anx = max([x.stop_utc for x in burst_infos])

        self.inputs = [
            get_subxml_from_metadata(path, metadata_type, self.swath, self.pol) for path in self.metadata_paths
        ]

        products = [get_subxml_from_metadata(path, 'product', self.swath, self.pol) for path in self.metadata_paths]
        slc_lengths = []
        for annotation in products:
            n_bursts = int(annotation.find('.//burstList').get('count'))
            burst_length = int(annotation.find('.//linesPerBurst').text)
            slc_lengths.append(n_bursts * burst_length)
        self.slc_lengths = slc_lengths

        # annotation components to be extended by subclasses
        self.ads_header = None
        self.xml = None

        # these attributes are updated when the annotation is written to a file
        self.size_bytes = None
        self.md5 = None

    def create_ads_header(self):
        """Create the ADS header for the annotation."""
        ads_header = deepcopy(self.inputs[0].find('adsHeader'))
        ads_header.find('startTime').text = self.min_anx.isoformat()
        ads_header.find('stopTime').text = self.max_anx.isoformat()
        ads_header.find('imageNumber').text = f'{self.image_number:03d}'
        self.ads_header = ads_header

    def merge_lists(self, list_name: str, line_bounds: Optional[tuple[int, int]] = None) -> ET.Element:
        """Merge lists of elements into a single list.

        Args:
            list_name: The name of the list element.

        Returns:
            The merged list element.
        """
        list_elements = [input_xml.find(list_name) for input_xml in self.inputs]
        list_of_list_elements = ListOfListElements(list_elements, self.start_line, self.slc_lengths)
        merged_list = list_of_list_elements.create_filtered_list([self.min_anx, self.max_anx], line_bounds=line_bounds)
        return merged_list

    def write(self, out_path: Path, update_info=True) -> None:
        """Write the annotation to a file.

        Args:
            out_path: The path to write the annotation to.
            update_info: Whether to update the size and md5 attributes of the annotation.
        """
        self.xml.write(out_path, pretty_print=True, xml_declaration=True, encoding='utf-8')

        if update_info:
            self.path = out_path
            with open(out_path, 'rb') as f:
                file_bytes = f.read()
                self.size_bytes = len(file_bytes)
                self.md5 = hashlib.md5(file_bytes).hexdigest()

    def __str__(self, **kwargs):
        """Return the XML string representation of the annotation.

        Args:
            kwargs: Keyword arguments to pass to the lxml
        """
        xml_str = ET.tostring(self.xml, pretty_print=True, **kwargs)
        return xml_str.decode()

    def create_manifest_components(self):
        """Create the components for the manifest file."""
        unit_type = 'Metadata Unit'
        mime_type = 'text/xml'

        rep_id = f's1Level1{self.metadata_type.capitalize()}Schema'
        simple_name = self.path.with_suffix('').name.replace('-', '')
        if self.metadata_type == 'product':
            simple_name = f'product{simple_name}'

        safe_index = [i for i, x in enumerate(self.path.parts) if 'SAFE' in x][-1]
        safe_path = Path(*self.path.parts[: safe_index + 1])
        rel_path = self.path.relative_to(safe_path)

        content_unit = create_content_unit(simple_name, unit_type, rep_id)
        metadata_object = create_metadata_object(simple_name)
        data_object = create_data_object(simple_name, rel_path, rep_id, mime_type, self.size_bytes, self.md5)

        return content_unit, metadata_object, data_object



from collections import namedtuple
from unittest.mock import patch

import pytest

from burst2safe import search


@pytest.fixture()
def product():
    properties = {
        'burst': {
            'relativeBurstID': 10,
            'fullBurstID': '001_000010_IW1',
        },
        'polarization': 'VV',
        'orbit': '000001',
    }
    BurstProduct = namedtuple('BurstProduct', ['properties'])
    return BurstProduct(properties=properties)


def test_add_surrounding_bursts(product):
    with patch('burst2safe.search.asf_search.search') as mock_search:
        search.add_surrounding_bursts([product], 3)
        burst_ids = ['001_000009_IW1', '001_000010_IW1', '001_000011_IW1']
        mock_search.assert_called_once_with(
            dataset='SLC-BURST', absoluteOrbit=1, polarization='VV', fullBurstID=burst_ids
        )

    with patch('burst2safe.search.asf_search.search') as mock_search:
        search.add_surrounding_bursts([product], 4)
        burst_ids = ['001_000009_IW1', '001_000010_IW1', '001_000011_IW1', '001_000012_IW1']
        mock_search.assert_called_once_with(
            dataset='SLC-BURST', absoluteOrbit=1, polarization='VV', fullBurstID=burst_ids
        )



"""Soucre: https://github.com/asfadmin/rain/blob/devel/tests/test_util_burst.py"""

from pytest import raises

from burst2safe import burst_id


CORRECT_GRANULE_NAME = 'S1_032322_IW2_20230422T170310_VV_ABCD'

EQUATORCROSSER = {
    'inputs': {
        'anx_time': '2023-04-22T17:03:10.235250',
        'azimuthAnxTime': float('6.208265985771300e+03'),
        'sensingTime': '2023-04-22T18:46:39.515927',
        'orbit_number_abs_start': 48213,
        'orbit_number_abs_stop': 48213,
        'orbit_number_rel_start': 16,
        'orbit_number_rel_stop': 16,
        'subswath': 'IW2',
    },
    'results': {
        'burstid_abs': 103556001,
        'burstid_rel': 32322,
        'orbit_num_abs': 48213,
        'orbit_num_rel': 16,
    },
}
EW = {
    'inputs': {
        'anx_time': '2022-10-10T14:02:11.848637',
        'azimuthAnxTime': float('1.816083970564404e+03'),
        'sensingTime': '2022-10-10T14:32:29.345783',
        'orbit_number_abs_start': 45381,
        'orbit_number_abs_stop': 45381,
        'orbit_number_rel_start': 159,
        'orbit_number_rel_stop': 159,
        'subswath': 'EW5',
    },
    'results': {
        'burstid_abs': 88487688,
        'burstid_rel': 308684,
        'orbit_num_abs': 45381,
        'orbit_num_rel': 159,
    },
}


def test_get_mode_timing():
    d = burst_id._get_mode_timing('EW1')
    assert d[0] == burst_id.PREAMBLE_LENGTH_EW
    assert d[1] == burst_id.BEAM_CYCLE_TIME_EW
    d = burst_id._get_mode_timing('IW2')
    assert d[0] == burst_id.PREAMBLE_LENGTH_IW
    assert d[1] == burst_id.BEAM_CYCLE_TIME_IW


def test_get_mode_timing_fail():
    with raises(burst_id.InvalidModeNameError):
        burst_id._get_mode_timing('NO')


def burstid_calc(vals: dict):
    bid, orbitnum = burst_id.calculate_burstid(
        vals['inputs']['sensingTime'],
        vals['inputs']['anx_time'],
        vals['inputs']['orbit_number_abs_start'],
        vals['inputs']['orbit_number_abs_stop'],
        vals['inputs']['subswath'],
    )
    assert bid == vals['results']['burstid_abs']
    assert orbitnum == vals['results']['orbit_num_abs']

    bid, orbitnum = burst_id.calculate_burstid(
        vals['inputs']['sensingTime'],
        vals['inputs']['anx_time'],
        vals['inputs']['orbit_number_rel_start'],
        vals['inputs']['orbit_number_rel_stop'],
        vals['inputs']['subswath'],
    )
    assert bid == vals['results']['burstid_rel']
    assert orbitnum == vals['results']['orbit_num_rel']


def test_calculate_burstid_opera_eqcross():
    burstid_calc(EQUATORCROSSER)


def test_calculate_burstid_opera_ew():
    burstid_calc(EW)


def test_calculate_burstid_opera_badsubsw():
    bad = EQUATORCROSSER.copy()
    bad['inputs']['subswath'] = 'NO8'
    with raises(burst_id.InvalidModeNameError):
        burstid_calc(bad)



from copy import deepcopy

from burst2safe import local2safe


def test_burst_from_local(burst_info1, burst_info2):
    burst_info1_copy = deepcopy(burst_info1)
    burst_info1_local = local2safe.burst_info_from_local(
        burst_info1.data_path,
        burst_info1.metadata_path,
        burst_info1.slc_granule,
        burst_info1.swath,
        burst_info1.polarization,
        burst_info1.burst_index,
    )
    assert burst_info1_local == burst_info1_copy

    burst_info2_copy = deepcopy(burst_info2)
    burst_info2_local = local2safe.burst_info_from_local(
        burst_info2.data_path,
        burst_info2.metadata_path,
        burst_info2.slc_granule,
        burst_info2.swath,
        burst_info2.polarization,
        burst_info2.burst_index,
    )
    assert burst_info2_local == burst_info2_copy


def test_load_burst_infos(burst_info1, burst_info2):
    slc_dict = {
        burst_info1.slc_granule: {
            burst_info1.swath: {
                burst_info1.polarization: {
                    str(burst_info1.burst_index): {
                        'DATA': burst_info1.data_path,
                        'METADATA': burst_info1.metadata_path,
                    },
                },
            }
        },
        burst_info2.slc_granule: {
            burst_info2.swath: {
                burst_info2.polarization: {
                    str(burst_info2.burst_index): {
                        'DATA': burst_info2.data_path,
                        'METADATA': burst_info2.metadata_path,
                    },
                },
            }
        },
    }
    burst_infos = local2safe.load_burst_infos(slc_dict)
    assert burst_infos[0] == burst_info1
    assert burst_infos[1] == burst_info2



"""Generate a SAFE file from local burst extractor outputs"""

import argparse
import json
from datetime import datetime
from pathlib import Path
from typing import Optional

from burst2safe import utils
from burst2safe.burst_id import calculate_burstid
from burst2safe.safe import Safe


def burst_info_from_local(
    tiff_path: Path, xml_path: Path, slc_name: str, swath: str, polarization: str, burst_index: int
) -> utils.BurstInfo:
    """Create a BurstInfo object from a local copy of the burst extractor output

    Args:
        xml_path: The path to the XML file
        swath: The name of the swath
        burst_index: The index of the burst within the swath
    """
    extractor_url = 'https://sentinel1-burst.asf.alaska.edu'
    burst_url_base = f'{extractor_url}/{slc_name}/{swath}/{polarization}/{burst_index}'
    data_url = f'{burst_url_base}.tiff'
    metadata_url = f'{burst_url_base}.xml'

    manifest = utils.get_subxml_from_metadata(xml_path, 'manifest', swath, polarization)
    xml_orbit_path = './/{*}metadataObject[@ID="measurementOrbitReference"]/metadataWrap/xmlData/{*}orbitReference'
    meta_orbit = manifest.find(xml_orbit_path)
    abs_orbit_start, abs_orbit_stop = [int(x.text) for x in meta_orbit.findall('{*}orbitNumber')]
    rel_orbit_start, rel_orbit_stop = [int(x.text) for x in meta_orbit.findall('{*}relativeOrbitNumber')]
    direction = meta_orbit.find('{*}extension/{*}orbitProperties/{*}pass').text.upper()

    product = utils.get_subxml_from_metadata(xml_path, 'product', swath, polarization)
    sensing_time_str = product.findall('swathTiming/burstList/burst')[burst_index].find('sensingTime').text
    anx_time_str = meta_orbit.find('{*}extension/{*}orbitProperties/{*}ascendingNodeTime').text
    burst_id, rel_orbit = calculate_burstid(sensing_time_str, anx_time_str, rel_orbit_start, rel_orbit_stop, swath)
    info = utils.BurstInfo(
        granule='',
        slc_granule=slc_name,
        swath=swath,
        polarization=polarization,
        burst_id=burst_id,
        burst_index=burst_index,
        direction=direction,
        absolute_orbit=abs_orbit_start,
        relative_orbit=rel_orbit_start,
        date=None,
        data_url=data_url,
        data_path=tiff_path,
        metadata_url=metadata_url,
        metadata_path=xml_path,
    )
    info.add_shape_info()
    info.add_start_stop_utc()
    date_format = '%Y%m%dT%H%M%S'
    start_utc_str = datetime.strftime(info.start_utc, date_format)
    info.date = datetime.strptime(datetime.strftime(info.start_utc, date_format), date_format)
    info.granule = f'S1_{burst_id}_{swath}_{start_utc_str}_{polarization}_{slc_name.split("_")[-1]}-BURST'
    return info


def load_burst_infos(slc_dict: dict) -> list[utils.BurstInfo]:
    """Load the burst infos from the SLC tree

    Args:
        slc_dict: SLC tree dict with format: {slc_name: {swath: {polarization: {burst_index: {DATA:tiff_path, METADATA:xml_path}}}}}
                Can contain an abitrary number of bursts as long as they are a valid merge group

    Returns:
        A list of BurstInfo objects
    """
    valid_swaths = ['IW1', 'IW2', 'IW3', 'EW1', 'EW2', 'EW3', 'EW4', 'EW5']
    valid_pols = ['VV', 'VH', 'HV', 'HH']
    burst_infos = []
    for slc_name in slc_dict:
        slc_name = slc_name.upper()
        swath_dict = slc_dict[slc_name]

        for swath in swath_dict:
            swath = swath.upper()
            if swath not in valid_swaths:
                raise ValueError(f'Invalid swath: {swath}')

            polarization_dict = swath_dict[swath]

            for polarization in polarization_dict:
                polarization = polarization.upper()
                if polarization not in valid_pols:
                    raise ValueError(f'Invalid polarization: {polarization}')

                burst_dict = polarization_dict[polarization]
                for burst_index in burst_dict:
                    burst_info = burst_info_from_local(
                        Path(burst_dict[burst_index]['DATA']),
                        Path(burst_dict[burst_index]['METADATA']),
                        slc_name,
                        swath,
                        polarization,
                        int(burst_index),
                    )
                    burst_infos.append(burst_info)

    return burst_infos


def local2safe(
    slc_dict: dict,
    all_anns: bool = False,
    keep_files: bool = False,
    work_dir: Optional[Path] = None,
) -> Path:
    """Convert a set of burst granules to the ESA SAFE format using local files

    Args:
        slc_dict: SLC tree dict with format: {slc_name: {swath: {polarization: {burst_index: {DATA:tiff_path, METADATA:xml_path}}}}}
                  Can contain an abitrary number of bursts as long as they are a valid merge group
        all_anns: Include product annotation files for all swaths, regardless of included bursts
        keep_files: Keep the intermediate files
        work_dir: The directory to store temporary files

    Returns:
        The path to the created SAFE
    """
    work_dir = utils.optional_wd(work_dir)

    burst_infos = load_burst_infos(slc_dict)
    print(f'Found {len(burst_infos)} burst(s).')

    print('Check burst group validity...')
    Safe.check_group_validity(burst_infos)
    print('Creating SAFE...')

    safe = Safe(burst_infos, all_anns, work_dir)
    safe_path = safe.create_safe()
    print('SAFE created!')

    if not keep_files:
        safe.cleanup()

    return safe_path


def main():
    """Entrypoint for the local2safe script
    Example:

    local2safe /path/to/slc_tree.json --all_anns --work_dir /path/to/work_dir
    """
    parser = argparse.ArgumentParser(description='Generate a SAFE file from local burst extractor outputs')
    parser.add_argument(
        'json_tree_path',
        type=Path,
        help='Path to the SLC tree JSON file with format: {slc_name: {swath: {polarization: {burst_index: {DATA:tiff_path, METADATA:xml_path}}}}}',
    )
    parser.add_argument('--all_anns', action='store_true', help='Include all annotations')
    parser.add_argument('--work_dir', type=Path, help='The directory to store temporary files')
    args = parser.parse_args()
    slc_tree = json.loads(args.json_tree_path.read_text())

    local2safe(
        slc_tree,
        args.all_anns,
        args.work_dir,
    )
